from __future__ import annotations

from prettyqt.qt.QtDesigner import *  # noqa: F403

from prettyqt import qt

if qt.API == "pyqt6":
    from .pydesignercustomwidgetcollectionplugin import (
        PyDesignerCustomWidgetCollectionPlugin,
    )
    from .pydesignercustomwidgetplugin import PyDesignerCustomWidgetPlugin
else:
    from .pydesignercustomwidgetcollection import PyDesignerCustomWidgetCollection

from .abstractextensionfactory import AbstractExtensionFactory
from .pydesignertaskmenuextension import PyDesignerTaskMenuExtension
from .designercustomwidgetinterface import DesignerCustomWidgetInterface
from .designerformeditorinterface import DesignerFormEditorInterface
from .abstractformbuilder import AbstractFormBuilder
from .formbuilder import FormBuilder
from prettyqt.qt import QtDesigner

QT_MODULE = QtDesigner

__all__ = [
    "AbstractExtensionFactory",
    "AbstractFormBuilder",
    "DesignerCustomWidgetInterface",
    "DesignerFormEditorInterface",
    "FormBuilder",
    "PyDesignerCustomWidgetCollection",
    "PyDesignerCustomWidgetCollectionPlugin",
    "PyDesignerCustomWidgetPlugin",
    "PyDesignerTaskMenuExtension",
]



from __future__ import annotations

import collections
import logging
from typing import TYPE_CHECKING, Literal, overload

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict, listdelegators


if TYPE_CHECKING:
    from collections.abc import Sequence


DOCK_OPTION = bidict(
    animated_docks=widgets.QMainWindow.DockOption.AnimatedDocks,
    allow_nested_docks=widgets.QMainWindow.DockOption.AllowNestedDocks,
    allow_tabbed_docks=widgets.QMainWindow.DockOption.AllowTabbedDocks,
    force_tabbed_docks=widgets.QMainWindow.DockOption.ForceTabbedDocks,
    vertical_tabs=widgets.QMainWindow.DockOption.VerticalTabs,
    grouped_dragging=widgets.QMainWindow.DockOption.GroupedDragging,
)

DockOptionStr = Literal[
    "animated_docks",
    "allow_nested_docks",
    "allow_tabbed_docks",
    "force_tabbed_docks",
    "vertical_tabs",
    "grouped_dragging",
]

DEFAULT_OPTS = (
    widgets.QMainWindow.DockOption.AllowTabbedDocks
    | widgets.QMainWindow.DockOption.AllowNestedDocks
    | widgets.QMainWindow.DockOption.GroupedDragging
    | widgets.QMainWindow.DockOption.AnimatedDocks
)

logger = logging.getLogger(__name__)


class PopupMenuAction(gui.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.triggered.connect(self._show_menu)
        self._mainwindow: MainWindow = self.parent()  # type: ignore

    def _show_menu(self):
        menu = self._mainwindow.createPopupMenu()
        self.setMenu(menu)
        menu.exec(gui.Cursor.pos())


class FullScreenAction(gui.Action):
    def __init__(
        self,
        parent: widgets.MainWindow,
        text: str = "Fullscreen",
        shortcut: str = "F11",
    ):
        super().__init__(
            text=text,
            # icon="mdi.fullscreen",
            checkable=True,
            shortcut=shortcut,
            parent=parent,
            checked=parent.isFullScreen(),
        )
        self.triggered.connect(parent.toggle_fullscreen)
        parent.installEventFilter(self)

    def eventFilter(self, source, event):
        """Needed to adjust checkstate when windowstate changes programatically."""
        if event.type() == core.Event.Type.WindowStateChange:
            self.setChecked(source.isFullScreen())
        return False


class MainWindow(widgets.WidgetMixin, widgets.QMainWindow):
    """Main application window."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dock_options=DEFAULT_OPTS, **kwargs)
        self.setMenuBar(widgets.MenuBar())

    def __getitem__(self, index: str) -> widgets.QWidget:
        result = self.find_child(widgets.QWidget, index)
        if result is None:
            msg = "Widget not found"
            raise KeyError(msg)
        return result

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "toolButtonStyle": constants.TOOLBUTTON_STYLE,
            "tabShape": widgets.tabwidget.TAB_SHAPES,
        }
        return maps

    def set_widget(self, widget: widgets.QWidget | None) -> widgets.QWidget | None:
        """Set widget and return previous one if existing."""
        previous = self.takeCentralWidget()
        if widget:
            self.setCentralWidget(widget)
        return previous

    def createPopupMenu(self) -> widgets.Menu:
        menu = widgets.Menu(parent=self, title="Window")
        for i, item in enumerate(self.get_docks()):
            action = gui.Action(
                text=item.windowTitle(),
                parent=self,
                checkable=True,
                checked=item.isVisible(),
                shortcut=f"Ctrl+Shift+{i}",
                shortcut_context="application",
                toggled=item.setVisible,
            )
            menu.add(action)
        menu.add_separator()
        for tb in self.get_toolbars():
            action = gui.Action(
                text=tb.windowTitle(),
                parent=self,
                checkable=True,
                toggled=tb.setVisible,
                checked=tb.isVisible(),
            )
            menu.add(action)
        return menu

    def add(self, widget: widgets.QWidget, **kwargs):
        match widget:
            case widgets.QToolBar():
                self.add_toolbar(widget, **kwargs)
            case widgets.QDockWidget():
                self.add_dockwidget(widget, **kwargs)
            case widgets.QWidget():
                self.centralWidget().layout().add(widget, **kwargs)

    def get_corner(
        self, corner: constants.CornerStr | constants.Corner
    ) -> constants.DockWidgetAreaStr:
        corner_flag = constants.CORNER.get_enum_value(corner)
        area = self.corner(corner_flag)
        return constants.DOCK_WIDGET_AREAS.inverse[area]

    def set_corner(
        self,
        corner: constants.CornerStr | constants.Corner,
        area: constants.DockWidgetAreaStr | constants.DockWidgetArea,
    ):
        corner_flag = constants.CORNER.get_enum_value(corner)
        area_flag = constants.DOCK_WIDGET_AREAS.get_enum_value(area)
        self.setCorner(corner_flag, area_flag)

    def add_toolbar(
        self,
        toolbar: widgets.QToolBar,
        area: constants.ToolbarAreaStr | constants.ToolBarArea | Literal["auto"] = "auto",
    ):
        """Adds a toolbar to the mainmenu at specified area.

        Args:
            toolbar: toolbar to use
            area: area of the toolbar
        """
        if area == "auto":
            area = self._get_preferred_toolbar_position()
        self.addToolBar(constants.TOOLBAR_AREA.get_enum_value(area), toolbar)

    def add_toolbar_break(
        self, position: constants.ToolbarAreaStr | constants.ToolBarArea = "top"
    ):
        """Adds a toolbar break to the given area behind the last item.

        Args:
            position: position of the toolbar
        """
        self.addToolBarBreak(constants.TOOLBAR_AREA.get_enum_value(position))

    def load_window_state(self, recursive: bool = False) -> bool:
        settings = core.Settings()
        name = self.objectName()
        geom = settings.get(f"{name}.geometry")
        state = settings.get(f"{name}.state")
        restored = False
        if geom is not None and state is not None:
            try:
                logger.debug("Loading window state for %r...", self.windowTitle())
                self.restoreGeometry(geom)
                if isinstance(state, str):
                    state = state.encode()
                self.restoreState(state)
                restored = True
            except TypeError:
                logger.exception(
                    "Wrong type for window state. Probably Qt binding switch?"
                )
        if recursive:
            for window in self.find_children(MainWindow, recursive=True):
                if window.objectName():
                    window.load_window_state()
        return restored

    def save_window_state(self, recursive: bool = False):
        """Save current window state as QSetting.

        Args:
            recursive (bool, optional): Description
        """
        settings = core.Settings()
        name = self.objectName()
        logger.debug("Saving window state for %r...", self.windowTitle())
        settings[f"{name}.geometry"] = self.saveGeometry()
        settings[f"{name}.state"] = self.saveState()
        if recursive:
            for window in self.find_children(MainWindow, recursive=True):
                if window.objectName():
                    window.save_window_state()

    def add_widget_as_dock(
        self,
        name: str,
        title: str,
        layout: widgets.layout.LayoutTypeStr | widgets.QLayout = "horizontal",
        position: constants.DockWidgetAreaStr | constants.DockWidgetArea = "left",
    ) -> widgets.DockWidget:
        dock_widget = widgets.DockWidget(self, object_name=name, window_title=title)
        widget = widgets.Widget()
        widget.setObjectName(f"{name}.widget")
        widget.set_layout(layout, margin=0)
        dock_widget.setWidget(widget)
        self.add_dockwidget(dock_widget, position)
        return dock_widget

    @overload
    def add_dockwidget(
        self,
        widget: widgets.QDockWidget,
        position: constants.DockWidgetAreaStr
        | constants.DockWidgetArea
        | Literal["auto"] = "auto",
    ) -> None: ...

    @overload
    def add_dockwidget(
        self,
        widget: widgets.QWidget,
        position: constants.DockWidgetAreaStr
        | constants.DockWidgetArea
        | Literal["auto"] = "auto",
        **kwargs,
    ) -> widgets.DockWidget: ...

    # Then keep your original implementation unchanged
    def add_dockwidget(
        self,
        widget: widgets.QWidget,
        position: (
            constants.DockWidgetAreaStr | constants.DockWidgetArea | Literal["auto"]
        ) = "auto",
        **kwargs,
    ) -> widgets.DockWidget | None:
        if position == "auto":
            position = self._get_preferred_dock_position()
        if not isinstance(widget, widgets.QDockWidget):
            dock_widget = widgets.DockWidget(self, **kwargs)
            dock_widget.set_widget(widget)
            self.addDockWidget(
                constants.DOCK_WIDGET_AREA.get_enum_value(position), dock_widget
            )
            return dock_widget
        self.addDockWidget(constants.DOCK_WIDGET_AREA.get_enum_value(position), widget)
        return None

    def remove(
        self,
        items: (
            Sequence[widgets.QDockWidget | widgets.QToolBar | gui.QAction]
            | widgets.QDockWidget
            | widgets.QToolBar
            | gui.QAction
        ),
    ):
        widget_list = items if isinstance(items, list) else [items]
        for i in widget_list:
            match i:
                case widgets.QDockWidget():
                    self.removeDockWidget(i)
                case widgets.QToolBar():
                    self.removeToolBar(i)
                case gui.QAction():
                    self.removeAction(i)

    def show_blocking(self):
        self.set_modality("application")
        self.show()

    def get_dock_area(self, widget: widgets.QDockWidget) -> constants.DockWidgetAreaStr:
        area = self.dockWidgetArea(widget)
        return constants.DOCK_WIDGET_AREAS.inverse[area]

    def get_toolbar_area(self, widget: widgets.QToolBar) -> constants.ToolbarAreaStr:
        area = self.toolBarArea(widget)
        return constants.TOOLBAR_AREA.inverse[area]

    def set_tool_button_style(
        self, style: constants.ToolButtonStyleStr | constants.ToolButtonStyle
    ):
        self.setToolButtonStyle(constants.TOOLBUTTON_STYLE.get_enum_value(style))

    def get_tool_button_style(self) -> constants.ToolButtonStyleStr:
        """Return current tool button style.

        Returns:
            tool button style
        """
        return constants.TOOLBUTTON_STYLE.inverse[self.toolButtonStyle()]

    def set_tab_shape(
        self, shape: widgets.tabwidget.TabShapeStr | widgets.QTabWidget.TabShape
    ):
        """Set tab shape for the tabwidget.

        Args:
            shape: tab shape to use
        """
        self.setTabShape(widgets.tabwidget.TAB_SHAPES.get_enum_value(shape))

    def get_tab_shape(self) -> widgets.tabwidget.TabShapeStr:
        """Return tab shape.

        Returns:
            tab shape
        """
        return widgets.tabwidget.TAB_SHAPES.inverse[self.tabShape()]

    def get_docks(
        self, position: constants.DockWidgetAreaStr | None = None
    ) -> listdelegators.ListDelegator[widgets.QDockWidget]:
        docks = self.find_children(widgets.QDockWidget, recursive=False)
        if position is None:
            return docks
        return listdelegators.ListDelegator([
            i for i in docks if self.get_dock_area(i) == position
        ])

    def _get_preferred_dock_position(
        self,
        preference: constants.DockWidgetAreaStr = "left",
    ) -> constants.DockWidgetAreaStr:
        """Get location with least amount of docks. If same score, use preference."""
        areas = [self.get_dock_area(i) for i in self.get_docks()]
        # by prepending the prio order, we can choose order because
        # collections.Counter takes insertion order into account.
        positions = ["bottom", "top", "right", "left"]
        positions.remove(preference)
        positions.append(preference)
        counter = collections.Counter(positions + areas)  # type: ignore[operator]
        return counter.most_common()[-1][0]  # type: ignore[return-value]

    def _get_preferred_toolbar_position(
        self,
        preference: constants.ToolbarAreaStr = "top",
    ) -> constants.ToolbarAreaStr:
        """See _get_preferred_dock_position."""
        areas = [self.get_toolbar_area(i) for i in self.get_toolbars()]
        positions = ["right", "left", "bottom", "top"]
        positions.remove(preference)
        positions.append(preference)
        counter = collections.Counter(positions + areas)  # type: ignore[operator]
        return counter.most_common()[-1][0]  # type: ignore[return-value]

    def get_toolbars(
        self, position: constants.ToolbarAreaStr | None = None
    ) -> listdelegators.ListDelegator[widgets.QToolBar]:
        toolbars = self.find_children(widgets.QToolBar, recursive=False)
        if position is None:
            return toolbars
        return [i for i in toolbars if self.get_toolbar_area(i) == position]


if __name__ == "__main__":
    app = widgets.app()
    form = MainWindow()
    dock = form.add_dockwidget(widgets.Widget())
    dock = form.add_dockwidget(widgets.Widget())
    dock = form.add_dockwidget(widgets.Widget())
    dock = form.add_dockwidget(widgets.Widget())
    action = FullScreenAction(parent=form)
    menu = widgets.Menu("test")
    menu.add_action(action)
    form.menuBar().add_menu(menu)
    form.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


ExclusionPolicyStr = Literal["none", "exclusive", "exclusive_optional"]

EXCLUSION_POLICY: bidict[ExclusionPolicyStr, QtGui.QActionGroup.ExclusionPolicy] = bidict(
    none=QtGui.QActionGroup.ExclusionPolicy.None_,
    exclusive=QtGui.QActionGroup.ExclusionPolicy.Exclusive,
    exclusive_optional=QtGui.QActionGroup.ExclusionPolicy.ExclusiveOptional,
)


class ActionGroup(core.ObjectMixin, QtGui.QActionGroup):
    """Groups actions together."""

    def __init__(self, parent: core.QObject | None = None, **kwargs):
        super().__init__(parent, **kwargs)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"exclusionPolicy": EXCLUSION_POLICY}
        return maps

    def __len__(self) -> int:
        return len(self.actions())

    def __getitem__(self, item: int) -> QtGui.QAction:
        return self.actions()[item]

    def set_exclusion_policy(
        self, policy: ExclusionPolicyStr | QtGui.QActionGroup.ExclusionPolicy | None
    ):
        """Set exclusion policy to use.

        Args:
            policy: exclusion policy to use
        """
        if policy is None:
            policy = "none"
        self.setExclusionPolicy(EXCLUSION_POLICY.get_enum_value(policy))

    def get_exclusion_policy(self) -> ExclusionPolicyStr:
        """Return current exclusion policy.

        Returns:
            exclusion policy
        """
        return EXCLUSION_POLICY.inverse[self.exclusionPolicy()]


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    action = ActionGroup()
    act = gui.Action()
    action.addAction(act)
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtGui


class UndoCommand(QtGui.QUndoCommand):
    def __len__(self) -> int:
        return self.childCount()

    def __getitem__(self, index: int) -> QtGui.QUndoCommand:
        if index >= self.childCount():
            raise IndexError(index)
        return self.child(index)

    # def __setstate__(self, state):
    #     for c in state["children"]:
    #         c.setParent(self)

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()



from __future__ import annotations

from prettyqt import constants, widgets


class DialogMixin(widgets.WidgetMixin):
    def __init__(
        self,
        *args,
        delete_on_close: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if delete_on_close:
            self.delete_on_close()

    def keyPressEvent(self, e):
        match e.key():
            case constants.Key.Key_Escape:
                self.close()
            case constants.Key.Key_F11 if self.isMaximized():
                self.showNormal()
            case constants.Key.Key_F11:
                self.showMaximized()
            case _:
                super().keyPressEvent(e)

    def delete_on_close(self):
        self.setAttribute(constants.WidgetAttribute.WA_DeleteOnClose)

    def add_buttonbox(self) -> widgets.DialogButtonBox:
        button_box = widgets.DialogButtonBox.create(ok=self.accept, cancel=self.reject)
        self.box.add(button_box)
        return button_box

    def show_blocking(self) -> bool:
        self.show()
        return bool(self.exec())

    def is_accepted(self) -> bool:
        return self.result() == widgets.QDialog.DialogCode.Accepted

    def main_loop(self) -> int:
        return self.exec()


class Dialog(DialogMixin, widgets.QDialog):
    """The base class of dialog windows."""


if __name__ == "__main__":
    import pathlib
    import pickle

    app = widgets.app()
    w = Dialog()
    path = pathlib.Path("data.pkl")
    with path.open("wb") as writer:
        pickle.dump(w, writer)
    with path.open("rb") as reader:
        w = pickle.load(reader)
    w.show_blocking()



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import get_repr


class RadialGradient(gui.GradientMixin, gui.QRadialGradient):
    """Used in combination with QBrush to specify a radial gradient brush."""

    def __repr__(self):
        return get_repr(
            self,
            self.get_center(),
            self.centerRadius(),
            self.get_focal_point(),
            self.focalRadius(),
        )

    def get_center(self) -> core.PointF:
        return core.PointF(self.center())

    def get_focal_point(self) -> core.PointF:
        return core.PointF(self.focalPoint())

    def get_css(self) -> str:
        """Convert gradient to a CSS string. Can be used for stylesheets."""
        cx, cy = self.center().x(), self.center().y()
        fx, fy = self.focalPoint().x(), self.focalPoint().y()
        stops = self.stops()
        stops = "\n".join(f"    stop: {stop:f} {color.name()}" for stop, color in stops)
        spread = self.get_spread()
        # couldnt find equivalent for focalRadius on first sight.
        return (
            "qradialgradient(\n"
            f"    cx: {cx}, cy: {cy}, radius: {self.radius()}, fx: {fx}, fy: {fy},\n"
            f"    spread:{spread},\n"
            f"{stops})"
        )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    grad = RadialGradient()
    grad.set_color_at(0, "red")
    grad.set_color_at(0.5, "blue")
    grad.set_color_at(1, "green")
    print(grad.get_css())
    btn = widgets.PushButton("TestButton")
    btn.setStyleSheet(f"QPushButton{{ background-color:{grad.get_css()} }}")
    btn.show()
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtCore


class ModelIndex(QtCore.QModelIndex):
    """Used to locate data in a data model."""

    def __getitem__(self, flag: int):
        return self.data(flag)



from __future__ import annotations

import html
import os
from typing import TYPE_CHECKING, Any

from prettyqt import constants, gui, iconprovider
from prettyqt.qt import QtCore, QtWidgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class TableWidgetItem(QtWidgets.QTableWidgetItem):
    """Item for use with the QTableWidget class."""

    def __setitem__(self, index: int | constants.ItemDataRoleStr, value):
        self.set_data(index, value)

    def __getitem__(self, index: int | constants.ItemDataRoleStr):
        return self.get_data(index)

    def set_flag(
        self, flag_name: constants.ItemFlagStr | constants.ItemFlag, value: bool
    ):
        """Set a flag based on str name."""
        flag = constants.ITEM_FLAG.get_enum_value(flag_name)
        if value:
            self.setFlags(self.flags() | flag)
        else:
            self.setFlags(self.flags() & ~flag)

    def set_editable(self, editable: bool):
        """Set whether this item is user-editable."""
        self.set_flag("editable", editable)

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the action.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def set_checkstate(self, state: constants.CheckStateStr | constants.CheckState):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use
        """
        self.setCheckState(constants.CHECK_STATE.get_enum_value(state))

    def get_checkstate(self) -> constants.CheckStateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState()]

    def set_text_alignment(
        self,
        horizontal: (
            constants.HorizontalAlignmentStr | constants.AlignmentFlag | None
        ) = None,
        vertical: constants.VerticalAlignmentStr | constants.AlignmentFlag | None = None,
    ):
        """Set text alignment of the checkbox.

        Args:
            horizontal: horizontal text alignment to use
            vertical: vertical text alignment to use
        """
        match horizontal, vertical:
            case None, None:
                return
            case None, _:
                flag = constants.V_ALIGNMENT.get_enum_value(vertical)
            case _, None:
                flag = constants.H_ALIGNMENT.get_enum_value(horizontal)
            case _, _:
                flag = constants.V_ALIGNMENT.get_enum_value(
                    vertical
                ) | constants.H_ALIGNMENT.get_enum_value(horizontal)
        self.setTextAlignment(flag)

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def get_foreground(self) -> gui.Brush:
        return gui.Brush(self.foreground())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_icon(self) -> gui.Icon | None:
        return None if (icon := self.icon()).isNull() else gui.Icon(icon)

    def set_data(self, role: constants.ItemDataRoleStr | int, data: Any):
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        super().setData(role, data)

    def get_data(self, role: constants.ItemDataRoleStr | int) -> Any:
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        return super().data(role)

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, QtCore.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)


if __name__ == "__main__":
    item = TableWidgetItem()



from __future__ import annotations

import logging

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class SliceChangeFlagsProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to selectively change the ItemFlags of the source model.

    === "Without proxy"

        ```py
        table = widgets.TreeView()
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        source_model = gui.StandardItemModel.from_dict(data)
        table.set_model(source_model)
        # table.proxifier.change_flags(enabled=False)
        ```
        <figure markdown>
          ![Image title](../../images/slicechangeflagsproxymodel_before.png)
        </figure>

    === "With proxy"

        ```py
        table = widgets.TreeView()
        data = dict(
            first=["John", "Mary"],
            last=["Doe", "Bo"],
            height=[5.5, 6.0],
            weight=[130, 150],
        )
        source_model = gui.StandardItemModel.from_dict(data)
        table.set_model(source_model)
        table.proxifier.change_flags(enabled=False)
        ```
        <figure markdown>
          ![Image title](../../images/slicechangeflagsproxymodel_after.png)
        </figure>


    Supports changing all ItemRoles by passing keyword arguments.

    Possible keyword arguments:

    * `selectable`: ItemFlag.ItemIsSelectable,
    * `editable`: ItemFlag.ItemIsEditable,
    * `drag_enabled`: ItemFlag.ItemIsDragEnabled,
    * `drop_enabled`: ItemFlag.ItemIsDropEnabled,
    * `user_checkable`: ItemFlag.ItemIsUserCheckable,
    * `enabled`: ItemFlag.ItemIsEnabled,
    * `auto_tristate`: ItemFlag.ItemIsAutoTristate,
    * `never_has_children`: ItemFlag.ItemNeverHasChildren,
    * `user_tristate`: ItemFlag.ItemIsUserTristate,



    ### Example

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[::2, 2:].change_flags(readable=False)
    table.show()
    ```

    or

    ```py
    indexer = (slice(None, None, 2), slice(2, None))
    proxy = itemmodels.SliceChangeFlagsProxyModel(indexer=indexer, readable=False)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "change_flags"
    ICON = "mdi.flag"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._flags_to_remove: constants.ItemFlag = constants.ItemFlag(0)
        self._flags_to_add: constants.ItemFlag = constants.ItemFlag(0)

    # def setData(
    #     self,
    #     index: core.ModelIndex,
    #     value: Any,
    #     role: constants.ItemDataRole = constants.EDIT_ROLE,
    # ) -> bool:
    #     if self.indexer.contains(index):
    #         logger.warning("Trying to set data on region covered by read-only proxy")
    #         return False
    #     return super().setData(index, value, role)

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        flags = super().flags(index)
        if self.indexer_contains(index):
            for flag in self._flags_to_remove:
                flags &= ~flag
            for flag in self._flags_to_add:
                flags |= flag
        return flags

    def set_flags_to_add(self, flags: constants.ItemFlag):
        with self.change_layout():
            self._flags_to_add = flags

    def get_flags_to_add(self) -> constants.ItemFlag:
        return self._flags_to_add

    def set_flags_to_remove(self, flags: constants.ItemFlag):
        with self.change_layout():
            self._flags_to_remove = flags

    def get_flags_to_remove(self) -> constants.ItemFlag:
        return self._flags_to_remove

    flags_to_add = core.Property(
        constants.ItemFlag,
        get_flags_to_add,
        set_flags_to_add,
        doc="Flags to set",
    )
    flags_to_remove = core.Property(
        constants.ItemFlag,
        get_flags_to_remove,
        set_flags_to_remove,
        doc="Flags to unset",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    data = dict(
        first=["John", "Mary"],
        last=["Doe", "Bo"],
        height=[5.5, 6.0],
        weight=[130, 150],
    )
    model = gui.StandardItemModel.from_dict(data)
    app = widgets.app()
    table = widgets.TableView()
    table.set_model(model)
    # proxy = table.proxifier[:, 2:].change_flags(enabled=False)
    table.show()
    table.resize(600, 150)
    table.h_header.resize_sections("stretch")
    table.set_title("Change flags")
    table.set_icon("mdi.flag")
    with app.debug_mode():
        app.exec()



# based on https://stackoverflow.com/questions/58891116/delegate-with-radio-buttons
# credits to musicamante

from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, gui, widgets


if TYPE_CHECKING:
    from collections.abc import Iterable


class RadioDelegate(widgets.StyledItemDelegate):
    ID = "radio"

    def __init__(self, parent: widgets.QWidget, items: Iterable[str]):
        super().__init__(parent)
        self.items = items
        self.choices: list[int | None] = [None for _ in self.items]

    def createEditor(
        self,
        parent: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ) -> widgets.Widget:
        editor = widgets.Widget(parent=parent, margin=0, auto_fill_background=True)
        # create a button group to keep track of the checked radio
        editor.button_group = widgets.ButtonGroup()
        # adding the widget as an argument to the layout constructor automatically
        # applies it to the widget
        layout = editor.set_layout("horizontal", margin=0)
        for i, k in enumerate(self.items):
            rb = widgets.RadioButton(k)
            layout.addWidget(rb)
            # prevent the radio to get focus from keyboard or mouse
            rb.set_focus_policy("none")
            rb.installEventFilter(self)
            editor.button_group.addButton(rb, i)
        # add a stretch to always align contents to the left
        layout.addStretch(1)

        # set a property that will be used for the mask
        editor.setProperty("offMask", gui.Region(editor.rect()))  # type: ignore
        editor.installEventFilter(self)
        return editor

    def eventFilter(self, source: core.QObject, event: core.QEvent) -> bool:
        match event.type():
            case core.Event.Type.MouseButtonPress:
                if isinstance(source, widgets.QRadioButton):
                    if not source.parent().hasFocus():
                        # the parent has no focus, set it and ignore the click
                        source.parent().setFocus()
                        return True
                elif not source.hasFocus():
                    # the container has been clicked, check
                    source.setFocus()
            case core.Event.Type.FocusIn:
                # event received as a consequence of setFocus
                # clear the mask to show it completely
                source.clearMask()
            case core.Event.Type.FocusOut:
                # another widget has requested focus, set the mask
                source.setMask(source.property("offMask"))
                # update the table viewport to get rid of possible
                # grid lines left after masking
                source.parent().update()
        return super().eventFilter(source, event)

    def updateEditorGeometry(
        self,
        editor: widgets.QWidget,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        rect = core.Rect(option.rect)
        min_width = editor.minimumSizeHint().width()
        if rect.width() < min_width:
            rect.setWidth(min_width)
        editor.setGeometry(rect)
        # create a new mask based on the option rectangle, then apply it
        mask = gui.Region(0, 0, option.rect.width(), option.rect.height())
        editor.setProperty("offMask", mask)  # type: ignore
        editor.setMask(mask)

    def setEditorData(self, editor: widgets.QWidget, index: core.ModelIndex):
        value = index.data(constants.DISPLAY_ROLE)  # type: ignore
        if value in self.items:
            editor.button_group.button(self.items.index(value)).setChecked(True)

    def setModelData(
        self,
        editor: widgets.QWidget,
        model: core.QAbstractItemModel,
        index: core.ModelIndex,
    ):
        button = editor.button_group.checkedId()
        if button >= 0:
            model.setData(
                index,
                self.items[button],
                constants.DISPLAY_ROLE,  # type: ignore
            )
            self.choices[button] = index.row()


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.TableWidget()
    widget.setColumnCount(3)
    widget.insertRow(0)
    widget.insertRow(0)
    widget.setHorizontalHeaderLabels(["LIB", "CELL", "area"])
    item = widgets.TableWidgetItem("test")
    widget[0, 0] = item
    widget[1, 1] = widgets.TableWidgetItem("test")
    widget[1, 2] = widgets.TableWidgetItem("test")
    widget[2, 1] = widgets.TableWidgetItem("test")
    delegate = RadioDelegate(widget, ["a", "b"])
    widget.setItemDelegateForColumn(0, delegate)
    widget.openPersistentEditor(item)
    widget.show()
    app.exec()



from __future__ import annotations

import contextlib
import logging
import re
from typing import TYPE_CHECKING, Literal

from prettyqt import core, gui, widgets
from prettyqt.utils import colors


if TYPE_CHECKING:
    from collections.abc import Iterator

    from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)

SEARCH_REGEX = re.compile(r"([^/]|\\/)+$")
SEARCH_FLAGS_REGEX = re.compile(r"([^/]|\\/)*?([^\\]/[biw]*)$")
REPLACE_REGEX = re.compile(
    r"""
    (?P<search>([^/]|\\/)*?[^\\])/(?P<replace>(([^/]|\\/)*[^\\])?)/(?P<flags>[abiw]*)$
""",
    re.VERBOSE,
)


class TextEditSelecter:
    def __init__(self, widget):
        self._widget = widget
        self._current_block = None
        self.search_buffer: str | None = None
        self.search_flags = None

    def __getitem__(self, index: int | slice):
        doc = self._widget.document()
        if isinstance(index, int):
            return doc.findBlockByNumber(index)
        start = doc.findBlockByNumber(index.start)
        end = doc.findBlockByNumber(index.stop)
        blocks = [start]
        while start != end and start.isValid():
            start = start.next()
            blocks.append(start)
        return blocks

    def goto_line(self, line_no: int, end_pos: Literal["top", "bottom"] | None = None):
        doc = self._widget.document()
        match end_pos:
            case "top":
                self._widget.moveCursor(gui.TextCursor.MoveOperation.End)
            case "bottom":
                self._widget.moveCursor(gui.TextCursor.MoveOperation.Start)
        lines = doc.blockCount()
        assert 1 <= line_no <= lines
        pos = doc.findBlockByLineNumber(line_no - 1).position()
        with self.current_cursor() as text_cursor:
            text_cursor.setPosition(pos)

    def get_selected_text(self) -> str:
        if self._widget.textCursor().hasSelection():
            return self._widget.textCursor().selectedText()
        return ""

    def highlight_current_line(self, color: datatypes.ColorType = None):
        if color is None:
            color = self._widget.get_palette().get_color("highlight")
        else:
            color = colors.get_color(color)
        extra_selections = []

        if not self._widget.isReadOnly():
            selection = widgets.TextEdit.ExtraSelection()
            selection.format.setBackground(color)
            prop = gui.QTextFormat.Property.FullWidthSelection
            selection.format.setProperty(prop, True)
            selection.cursor = self._widget.textCursor()
            selection.cursor.clearSelection()
            extra_selections.append(selection)

        self._widget.setExtraSelections(extra_selections)

    def select_text(self, start: int, end: int):
        with self.create_cursor() as c:
            c.select_text(start, end)

    def _update_on_block_change(self):
        tc = self._widget.textCursor()
        b = tc.blockNumber()
        if b == self._current_block:
            return
        self._current_block = b
        self._widget.viewport().update()

    def get_current_line(self) -> int:
        return self._widget.textCursor().blockNumber()

    def get_selected_block_span(self) -> tuple[int, int]:
        start = self._widget.textCursor().selectionStart()
        end = self._widget.textCursor().selectionEnd()
        start_block_id = self._widget.document().findBlock(start).blockNumber()
        end_block_id = self._widget.document().findBlock(end).blockNumber()

        return (start_block_id, end_block_id)

    @contextlib.contextmanager
    def create_cursor(self) -> Iterator[gui.TextCursor]:
        cursor = gui.TextCursor(self._widget.document())
        yield cursor
        self._widget.setTextCursor(cursor)

    @contextlib.contextmanager
    def current_cursor(self) -> Iterator[gui.TextCursor]:
        cursor = gui.TextCursor(self._widget.textCursor())
        yield cursor
        self._widget.setTextCursor(cursor)

    def get_text_cursor(self) -> gui.TextCursor:
        return gui.TextCursor(self._widget.textCursor())

    def move_cursor(
        self,
        operation: gui.textcursor.MoveOperationStr,
        mode: gui.textcursor.MoveModeStr = "move",
    ):
        self._widget.moveCursor(
            gui.textcursor.MOVE_OPERATION[operation], gui.textcursor.MOVE_MODE[mode]
        )

    def search_and_replace(self, arg: str):
        """Main search and replace function.

        arg is a string with a vim-like s&r syntax (see the regexes below)
        """

        def generate_flags(flagstr: str):
            self.search_flags = gui.QTextDocument.FindFlag(0)
            if "b" in flagstr:
                self.search_flags |= gui.QTextDocument.FindFlag.FindBackward
            if "i" not in flagstr:
                self.search_flags |= gui.QTextDocument.FindFlag.FindCaseSensitively
            if "w" in flagstr:
                self.search_flags |= gui.QTextDocument.FindFlag.FindWholeWords

        if search_match := SEARCH_REGEX.match(arg):
            self.search_buffer = search_match[0]
            self.search_flags = gui.QTextDocument.FindFlag.FindCaseSensitively
            self.search_next()
        elif search_flags_match := SEARCH_FLAGS_REGEX.match(arg):
            self.search_buffer, flags = search_flags_match[0].rsplit("/", 1)
            generate_flags(flags)
            self.search_next()
        elif replace_match := REPLACE_REGEX.match(arg):
            self.search_buffer = replace_match.group("search")
            generate_flags(replace_match.group("flags"))
            if "a" in replace_match.group("flags"):
                self._replace_all(replace_match.group("replace"))
            else:
                self._replace_next(replace_match.group("replace"))
        else:
            logger.error("Malformed search/replace expression")

    def _searching_backwards(self) -> bool:
        return bool(gui.QTextDocument.FindFlag.FindBackward & self.search_flags)

    def search_next(self):
        """Go to the next string found.

        This does the same thing as running the same search-command again.
        """
        if self.search_buffer is None:
            logger.error("No previous searches")
            return
        temp_cursor = self._widget.textCursor()
        if not self._widget.find(self.search_buffer, self.search_flags):
            if not self._widget.textCursor().atStart() or (
                self._searching_backwards() and not self._widget.textCursor().atEnd()
            ):
                self.move_cursor("end" if self._searching_backwards() else "start")
                if not self._widget.find(self.search_buffer, self.search_flags):
                    self._widget.setTextCursor(temp_cursor)
                    logger.error("Text not found")
            else:
                self._widget.setTextCursor(temp_cursor)
                logger.error("Text not found")

    def _replace_next(self, replace_buffer: str) -> None:
        """Go to the next string found and replace it with replace_buffer.

        While this technically can be called from outside this class, it is
        not recommended (and most likely needs some modifications of the code.)
        """
        if self.search_buffer is None:
            logger.error("No previous searches")
            return
        temp_cursor = self._widget.textCursor()
        found = self._widget.find(self.search_buffer, self.search_flags)
        if not found and (
            not self._widget.textCursor().atStart()
            or (self._searching_backwards() and not self._widget.textCursor().atEnd())
        ):
            self.move_cursor("end" if self._searching_backwards() else "start")
            found = self._widget.find(self.search_buffer, self.search_flags)
            if not found:
                self._widget.setTextCursor(temp_cursor)
        if found:
            t = self._widget.textCursor()
            t.insertText(replace_buffer)
            length = len(replace_buffer)
            t.setPosition(t.position() - length)
            t.setPosition(t.position() + length, gui.QTextCursor.MoveMode.KeepAnchor)
            self._widget.setTextCursor(t)
            logger.info(
                "Replaced on line %s, pos %s", t.blockNumber(), t.positionInBlock()
            )
        else:
            logger.error("Text not found")

    def _replace_all(self, replace_buffer: str) -> None:
        """Replace all strings found with the replace_buffer.

        As with replace_next, you probably don't want to call this manually.
        """
        if self.search_buffer is None:
            logger.error("No previous searches")
            return
        temp_cursor = self._widget.textCursor()
        times = 0
        self.move_cursor("start")
        while self._widget.find(self.search_buffer, self.search_flags):
            self._widget.textCursor().insertText(replace_buffer)
            times += 1
        if times:
            logger.info("%s instance %s replaced", times, "s" if times else "")
        else:
            logger.error("Text not found")
        self._widget.setTextCursor(temp_cursor)

    def highlight_matches(
        self, pattern: str, highlight_color: datatypes.ColorType = "red"
    ):
        color = colors.get_color(highlight_color)
        cursor = self._widget.get_text_cursor()
        fmt = gui.QTextCharFormat()
        fmt.setBackground(gui.QBrush(color))
        re = core.RegularExpression(pattern)
        for match in re.global_match(self._widget.toPlainText()):
            cursor.set_position(match.capturedStart(), "move")
            cursor.set_position(match.capturedEnd(), "keep")
            cursor.mergeCharFormat(fmt)

    def replace_block_at_cursor(self, new_text: str):
        cursor = self._widget.textCursor()
        cursor.select(gui.QTextCursor.MoveOperation.BlockUnderCursor)
        if cursor.selectionStart() != 0:
            new_text = "\n" + new_text
        cursor.removeSelectedText()
        cursor.insertText(new_text)


if __name__ == "__main__":
    app = widgets.app()

    test = widgets.PlainTextEdit()
    for i in range(200):
        test.append_text(str(i))
    test.show()
    with app.debug_mode():
        app.sleep(2)
        print(test.selecter[20:50])
        app.exec()



from __future__ import annotations

from prettyqt import constants, core


class BasicTimer(core.QBasicTimer):
    """Timer events for objects."""

    def __bool__(self):
        return self.isActive()

    def start_timer(
        self, msec: int, obj: core.QObject, timer_type: constants.TimerTypeStr
    ):
        self.start(msec, constants.TIMER_TYPE.get_enum_value(timer_type), obj)



from __future__ import annotations

import contextlib
import os
import pathlib
from typing import TYPE_CHECKING, Literal

import qstylizer.parser
import qstylizer.style

from prettyqt import constants, core, gui
from prettyqt.utils import bidict, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterator

    from prettyqt.utils import datatypes


MarkdownFeatureStr = Literal["no_html", "commonmark", "github"]

MARKDOWN_FEATURES: bidict[MarkdownFeatureStr, gui.QTextDocument.MarkdownFeature] = bidict(
    no_html=gui.QTextDocument.MarkdownFeature.MarkdownNoHTML,
    commonmark=gui.QTextDocument.MarkdownFeature.MarkdownDialectCommonMark,
    github=gui.QTextDocument.MarkdownFeature.MarkdownDialectGitHub,
)

ResourceTypeStr = Literal["unknown", "html", "image", "stylesheet", "markdown", "user"]

RESOURCE_TYPES: bidict[ResourceTypeStr, gui.QTextDocument.ResourceType] = bidict(
    unknown=gui.QTextDocument.ResourceType.UnknownResource,
    html=gui.QTextDocument.ResourceType.HtmlResource,
    image=gui.QTextDocument.ResourceType.ImageResource,
    stylesheet=gui.QTextDocument.ResourceType.StyleSheetResource,
    markdown=gui.QTextDocument.ResourceType.MarkdownResource,
    user=gui.QTextDocument.ResourceType.UserResource,
)

StackStr = Literal["undo", "redo", "undo_and_redo"]

STACKS: bidict[StackStr, gui.QTextDocument.Stacks] = bidict(
    undo=gui.QTextDocument.Stacks.UndoStack,
    redo=gui.QTextDocument.Stacks.RedoStack,
    undo_and_redo=gui.QTextDocument.Stacks.UndoAndRedoStacks,
)

FindFlagStr = Literal["backward", "case_sensitive", "whole_words"]

FIND_FLAGS: bidict[FindFlagStr, gui.QTextDocument.FindFlag] = bidict(
    backward=gui.QTextDocument.FindFlag.FindBackward,
    case_sensitive=gui.QTextDocument.FindFlag.FindCaseSensitively,
    whole_words=gui.QTextDocument.FindFlag.FindWholeWords,
)

MetaInformationStr = Literal["document_title", "document_url", "css_media"]

META_INFORMATION: bidict[MetaInformationStr, gui.QTextDocument.MetaInformation] = bidict(
    document_title=gui.QTextDocument.MetaInformation.DocumentTitle,
    document_url=gui.QTextDocument.MetaInformation.DocumentUrl,
    css_media=gui.QTextDocument.MetaInformation.CssMedia,
)


class TextDocumentMixin(core.ObjectMixin):
    def __getitem__(self, index: int) -> gui.TextBlock:
        return gui.TextBlock(self.findBlockByNumber(index))

    def __len__(self) -> int:
        return self.blockCount()

    def __iter__(self) -> Iterator[gui.TextBlock]:
        return iter(
            gui.TextBlock(self.findBlockByNumber(i)) for i in range(self.blockCount())
        )

    def __repr__(self):
        return get_repr(self, self.toPlainText())

    def get_first_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.firstBlock())

    def get_last_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.lastBlock())

    def find_block_by_number(self, number: int) -> gui.TextBlock:
        # starts with 0-index
        block = self.findBlockByNumber(number)
        if not block.isValid():
            msg = f"{number} not a valid block index. Block count: {self.blockCount()}"
            raise ValueError(msg)
        return gui.TextBlock(block)

    def find_block_by_line_number(self, line_number: int) -> gui.TextBlock:
        # starts with 0-index
        block = self.findBlockByLineNumber(line_number)
        if not block.isValid():
            msg = f"{line_number} not a valid line index. Line count: {self.lineCount()}"
            raise ValueError(msg)
        return gui.TextBlock(block)

    def set_text(self, text: str):
        self.setPlainText(text)

    def get_base_url(self) -> core.Url:
        return core.Url(self.baseUrl())

    def get_default_font(self) -> gui.Font:
        return gui.Font(self.defaultFont())

    def set_default_text_option(self, opt: gui.QTextOption):
        self.setDefaultTextOption(gui.TextOption(opt))

    def get_default_text_option(self) -> gui.TextOption:
        return gui.TextOption(self.defaultTextOption())

    def clear_stacks(self, stack: StackStr | gui.QTextDocument.Stacks):
        """Clear undo / redo stack.

        Args:
            stack: stack to clear
        """
        self.clearUndoRedoStacks(STACKS.get_enum_value(stack))

    def set_default_cursor_move_style(
        self, style: constants.CursorMoveStyleStr | constants.CursorMoveStyle
    ):
        """Set the cursor move style.

        Args:
            style: cursor move style
        """
        self.setDefaultCursorMoveStyle(constants.CURSOR_MOVE_STYLE.get_enum_value(style))

    def get_default_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return current cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.defaultCursorMoveStyle()]

    def set_meta_information(
        self, info: MetaInformationStr | gui.QTextDocument.MetaInformation, value: str
    ):
        """Set meta information.

        Args:
            info: meta information type
            value: value to set
        """
        self.setMetaInformation(META_INFORMATION.get_enum_value(info), value)

    def get_meta_information(
        self, info: MetaInformationStr | gui.QTextDocument.MetaInformation
    ) -> str:
        """Return specififed meta information.

        Args:
            info: meta information type

        Returns:
            meta information
        """
        return self.metaInformation(META_INFORMATION.get_enum_value(info))

    def add_resource(
        self,
        resource_type: ResourceTypeStr | gui.QTextDocument.ResourceType,
        name: datatypes.PathType,
        resource,
    ):
        url = core.Url(name)
        self.addResource(RESOURCE_TYPES.get_enum_value(resource_type), url, resource)

    @contextlib.contextmanager
    def edit_default_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_default_stylesheet()
        yield ss
        self.set_default_stylesheet(ss)

    def set_default_stylesheet(
        self, ss: None | qstylizer.style.StyleSheet | datatypes.PathType
    ):
        match ss:
            case os.PathLike():
                ss = pathlib.Path(ss).read_text("utf-8")
            case None:
                ss = ""
            case qstylizer.style.StyleSheet():
                pass
            case _:
                raise TypeError(ss)
        self.setDefaultStyleSheet(str(ss))

    def get_default_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.defaultStyleSheet())

    def find_line_position(self, line_no: int) -> int:
        """Return index of the TextBlock corresponding to given line number."""
        lines = self.blockCount()
        assert 1 <= line_no <= lines
        return self.findBlockByLineNumber(line_no - 1).position()

    def write_to_file(
        self,
        path: datatypes.PathType,
        fmt: gui.textdocumentwriter.FormatStr | bytes | core.QByteArray = "plaintext",
    ) -> bool:
        writer = gui.TextDocumentWriter()
        writer.set_format(fmt)
        writer.set_file_name(path)
        return writer.write(self)

    def get_bytes(
        self, fmt: Literal["plaintext", "HTML", "markdown", "ODF"] = "ODF"
    ) -> bytes:
        return gui.TextDocumentWriter.serialize_document(self, fmt)

    def show_whitespace_and_tabs(self, show: bool):
        """Set show white spaces flag."""
        options = self.get_default_text_option()
        flag = gui.QTextOption.Flag.ShowTabsAndSpaces
        if show:
            options.setFlags(options.flags() | flag)
        else:
            options.setFlags(options.flags() & ~flag)
        self.setDefaultTextOption(options)

    def get_pixel_height(self, exact: bool = True) -> int:
        if exact:  # sourcery skip: extract-method
            layout = self.documentLayout()
            h = 0
            b = self.begin()
            while b != self.end():
                h += layout.blockBoundingRect(b).height()
                b = b.next()
            h += self.documentMargin()
            return h
        line_count = self.document().blockCount()
        font_metrics = gui.FontMetrics(self.defaultFont())
        return (max(line_count, 1) * font_metrics.height()) + self.documentMargin()


class TextDocument(TextDocumentMixin, gui.QTextDocument):
    pass


if __name__ == "__main__":
    doc = TextDocument("This is a test\nHello")
    doc.set_default_text_option(gui.QTextOption())
    a = doc.get_bytes()
    print(a)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui


if TYPE_CHECKING:
    from collections.abc import Callable


class FunctionValidator(gui.Validator):
    """Validator which checks based on a given Callable.

    Allows for quickly creating a validator without subclassing.
    """

    ID = "function"

    def __init__(self, fn: Callable[[str], bool], parent=None):
        super().__init__(parent)
        self._fn = fn

    def __eq__(self, other: object):
        return isinstance(other, FunctionValidator) and other._fn == self._fn

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        if self._fn(text):
            return self.State.Acceptable, text, pos
        return self.State.Invalid, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = FunctionValidator(lambda text: text in ["a", "b"])
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.set_validator(val, strict=False)
    widget.show()
    print(widget.validator())
    with app.debug_mode():
        app.exec()
    print(widget.hasAcceptableInput())



from __future__ import annotations

from prettyqt.qt import QtGui


class PixmapCache(QtGui.QPixmapCache):
    """Application-wide cache for pixmaps."""

    def __setitem__(self, key: str, value: QtGui.QPixmap):
        self.insert(key, value)

    def __getitem__(self, key: str) -> QtGui.QPixmap | None:
        return self.find(key)

    def __delitem__(self, key: str):
        self.remove(key)


if __name__ == "__main__":
    from prettyqt import gui
    from prettyqt.qt import QtCore

    app = gui.app()
    cache = PixmapCache()
    pix = gui.Pixmap.create_char("a", 10)
    cache["test"] = pix
    test = cache["test"]

    ba = QtCore.QByteArray()
    stream = QtCore.QDataStream(ba, QtCore.QIODeviceBase.OpenModeFlag.WriteOnly)
    stream << test
    assert ba.data() == bytes(pix)



from __future__ import annotations

from prettyqt import constants, core, widgets


class TableWidgetMixin(widgets.TableViewMixin):
    def __getitem__(self, index: tuple[int, int]) -> widgets.QTableWidgetItem:
        item = self.item(*index)
        if item is None:
            raise KeyError(index)
        return item

    def __setitem__(self, index: tuple[int, int], value: widgets.QTableWidgetItem):
        self.setItem(index[0], index[1], value)

    def __delitem__(self, index: tuple[int, int]):
        self.takeItem(*index)

    def sort(self, column: int = 0, reverse: bool = False):
        order = constants.DESCENDING if reverse else constants.ASCENDING
        self.sortItems(column, order)

    def openPersistentEditor(self, index: core.ModelIndex | widgets.QTableWidgetItem):
        if isinstance(index, core.ModelIndex):
            index = self.itemFromIndex(index)
        super().openPersistentEditor(index)

    def closePersistentEditor(self, index: core.ModelIndex | widgets.QTableWidgetItem):
        if isinstance(index, core.ModelIndex):
            index = self.itemFromIndex(index)
        super().closePersistentEditor(index)

    def isPersistentEditorOpen(
        self, index: core.ModelIndex | widgets.QTableWidgetItem
    ) -> bool:
        if isinstance(index, core.ModelIndex):
            index = self.itemFromIndex(index)
        return super().isPersistentEditorOpen(index)

    def scroll_to_item(
        self,
        item: widgets.QTableWidgetItem,
        scroll_hint: widgets.abstractitemview.ScrollHintStr = "ensure_visible",
    ):
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT[scroll_hint])


class TableWidget(TableWidgetMixin, widgets.QTableWidget):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = TableWidget()
    widget.setHorizontalHeaderLabels(["testus"])
    widget.setColumnCount(1)
    widget.insertRow(0)
    widget[0, 0] = widgets.TableWidgetItem("test")
    widget.show()
    app.exec()



"""Provides QtCore classes and functions."""

from __future__ import annotations

import datetime

from prettyqt.qt import PYQT6, PYSIDE6
from prettyqt.utils import get_repr

# sys.setrecursionlimit(2000)


def to_datetime(self) -> datetime.datetime:
    return self.toPyDateTime()


def to_date(self) -> datetime.date:
    return self.toPyDate()


def to_time(self) -> datetime.time:
    return self.toPyTime()


if PYQT6:
    from PyQt6.QtCore import (
        QT_VERSION_STR,
        QT_VERSION_STR as __version__,
        QDateTime,
        QDate,
        QTime,
        pyqtBoundSignal,
        pyqtBoundSignal as SignalInstance,
        pyqtProperty,
        pyqtProperty as Property,
        pyqtSignal,
        pyqtSignal as Signal,
        pyqtSlot,
        pyqtSlot as Slot,
        pyqtEnum,
        pyqtEnum as QEnum,
        pyqtEnum as QFlag,
        pyqtClassInfo,
        pyqtClassInfo as ClassInfo,
        PYQT_VERSION_STR as BINDING_VERSION,
        QtMsgType,
    )

    QtCriticalMsg = QtMsgType.QtCriticalMsg
    QtDebugMsg = QtMsgType.QtDebugMsg
    QtFatalMsg = QtMsgType.QtFatalMsg
    QtInfoMsg = QtMsgType.QtInfoMsg
    QtSystemMsg = QtMsgType.QtSystemMsg
    QtWarningMsg = QtMsgType.QtWarningMsg

    # For issue #153
    from PyQt6.QtCore import *

    QDateTime.toPython = to_datetime
    QDate.toPython = to_date
    QTime.toPython = to_time
    QLibraryInfo.location = QLibraryInfo.path  # type: ignore
    # Those are imported from `import *`
    del (
        pyqtSignal,
        pyqtBoundSignal,
        pyqtSlot,
        pyqtProperty,
        QT_VERSION_STR,
        pyqtEnum,
        pyqtClassInfo,
    )
elif PYSIDE6:
    from PySide6.QtCore import *  # type: ignore
    from PySide6.QtCore import __version__  # type: ignore
    from PySide6 import __version__ as BINDING_VERSION


# def __repr__(self):
#     return get_repr(self, self.objectName())


# QObject.__repr__ = __repr__



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui
from prettyqt.utils import get_repr


if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from prettyqt.qt import QtCore


class CompositeValidator(gui.Validator):
    """BaseClass for combined validators."""

    ID = "composite"

    def __init__(
        self,
        validators: Sequence[gui.Validator] | None = None,
        parent: QtCore.QObject | None = None,
    ):
        super().__init__(parent)
        self.validators = list(validators) if validators else []

    def __repr__(self):
        return get_repr(self, self.validators)

    def __getitem__(self, index: int) -> gui.Validator:
        return self.validators[index]

    def __setitem__(self, index: int, value: gui.Validator):
        self.validators[index] = value

    def __delitem__(self, index: int):
        del self.validators[index]

    def __contains__(self, index: int):
        return index in self.validators

    def __iter__(self) -> Iterator[gui.Validator]:
        return iter(self.validators)

    def __reduce__(self):
        return type(self), (self.validators,)

    def __len__(self):
        return len(self.validators)

    def __eq__(self, other: object):
        return (
            self.validators == other.validators
            if isinstance(other, type(self))
            else False
        )

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        return NotImplemented


class AndValidator(CompositeValidator):
    """Validator to AND-combine other validators."""

    ID = "and"

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        vals = [v.validate(text, pos)[0] for v in self.validators]  # type: ignore
        if self.State.Invalid in vals:
            return self.State.Invalid, text, pos
        if self.State.Intermediate in vals:
            return self.State.Intermediate, text, pos
        return self.State.Acceptable, text, pos


class OrValidator(CompositeValidator):
    """Validator to OR-combine other validators."""

    ID = "or"

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        vals = [v.validate(text, pos)[0] for v in self.validators]  # type: ignore
        if self.State.Acceptable in vals:
            return self.State.Acceptable, text, pos
        if self.State.Intermediate in vals:
            return self.State.Intermediate, text, pos
        return self.State.Invalid, text, pos


if __name__ == "__main__":
    from prettyqt import validators, widgets

    val1 = validators.NotEmptyValidator()
    val2 = validators.PathValidator()
    val = AndValidator([val1, val2])
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import location


class PlaceContentRequest(location.QPlaceContentRequest):
    """Represents the parameters of a content request."""

    def set_content_type(
        self, typ: location.placecontent.TypeStr | location.PlaceContent.Type
    ):
        """Set the content type.

        Args:
            typ: Relevance type
        """
        self.setContentType(location.placecontent.TYPE.get_enum_value(typ))

    def get_content_type(self) -> location.placecontent.TypeStr:
        """Return current content type.

        Returns:
            Relevance type
        """
        return location.placecontent.TYPE.inverse[self.contentType()]


if __name__ == "__main__":
    request = PlaceContentRequest()



"""Provides QtHelp classes and functions."""

from prettyqt.qt import PYQT6, PYSIDE6


if PYQT6:
    from PyQt6.QtHelp import *
elif PYSIDE6:
    from PySide6.QtHelp import *
else:
    raise ModuleNotFoundError("No Qt bindings could be found")



from __future__ import annotations

from collections.abc import Sequence
import datetime
import logging
import traceback
from typing import ClassVar

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class LevelNameColumn(itemmodels.ColumnItem):
    name = "Level"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.levelname


class FileNameColumn(itemmodels.ColumnItem):
    name = "File name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.filename


class FunctionNameColumn(itemmodels.ColumnItem):
    name = "Function name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.funcName


class LineNoColumn(itemmodels.ColumnItem):
    name = "Line number"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return str(item.lineno)


class ModuleColumn(itemmodels.ColumnItem):
    name = "Module"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.module


class CreatedColumn(itemmodels.ColumnItem):
    name = "Created"
    display_format = "yyyy-MM-dd hh:mm:ss.zzzzzz"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                dt = core.DateTime.from_seconds(item.created)
                return dt.toString(self.display_format)
            case constants.USER_ROLE:
                return core.DateTime.from_seconds(item.created)


class ProcessColumn(itemmodels.ColumnItem):
    name = "Process"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return str(item.process)
            case constants.USER_ROLE:
                return item.process


class ThreadColumn(itemmodels.ColumnItem):
    name = "Thread"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return str(item.thread)
            case constants.USER_ROLE:
                return item.thread


class ProcessNameColumn(itemmodels.ColumnItem):
    name = "Process name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.processName or ""


class ThreadNameColumn(itemmodels.ColumnItem):
    name = "Thread name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.threadName or ""


class RelativeCreatedColumn(itemmodels.ColumnItem):
    name = "Relative created"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return str(datetime.timedelta(milliseconds=item.relativeCreated))


class NameColumn(itemmodels.ColumnItem):
    name = "Name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.name


class PathNameColumn(itemmodels.ColumnItem):
    name = "Path name"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.pathname


class MessageColumn(itemmodels.ColumnItem):
    name = "Message"

    def get_data(self, item: logging.LogRecord, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return (
                    traceback.format_exc()
                    if isinstance(item.msg, Exception)
                    else item.msg % item.args
                )


class LogRecordModel(itemmodels.ColumnTableModel):
    """Model to display a list of logging.LogRecords."""

    COLUMNS: ClassVar = [
        LevelNameColumn,
        FileNameColumn,
        FunctionNameColumn,
        LineNoColumn,
        MessageColumn,
        ModuleColumn,
        CreatedColumn,
        ProcessColumn,
        ThreadColumn,
        ProcessNameColumn,
        ThreadNameColumn,
        RelativeCreatedColumn,
        NameColumn,
        PathNameColumn,
    ]
    SUPPORTS = Sequence[logging.LogRecord]

    def __init__(self, **kwargs):
        super().__init__(items=[], columns=self.COLUMNS, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (logging.LogRecord(), *_):
                return True
            case _:
                return False


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    logger = logging.getLogger()

    app = widgets.app()
    with app.debug_mode():
        with debugging.Stalker(app) as stalker:
            stalker.show()
        app.exec()
        # def raise_exc():
        #     try:
        #         raise Exception("test")
        #     except Exception as e:
        #         logger.exception(e)



from __future__ import annotations

import logging

import mknodes
from mknodes.pages import processors

from prettyqt import core, gui, itemmodels, prettyqtmarkdown, widgets
from prettyqt.utils import classhelpers


SLICE_PROXY_INFO = """This is a [slice proxy](SliceIdentityProxyModel.md)
and can be selectively applied to a model. Read more about
 [slices](https://docs.python.org/3/library/functions.html#slice)."""

RECURSIVE_MODEL_INFO = "Model can be recursive, so be careful with iterating whole tree."

logger = logging.getLogger(__name__)
app = widgets.app()


class QtParentContainerProcessor(processors.ContainerProcessor):
    ID = "qt_signature"

    def append_block(self, page: mknodes.MkPage):
        qt_parent = classhelpers.get_qt_parent_class(self.item)
        assert qt_parent
        provider = page.ctx.links if page.ctx else prettyqtmarkdown.QtLinkProvider()
        page += f"Qt Base Class: {provider.link_for_klass(qt_parent)}"
        page += f"Signature: `{qt_parent.__doc__}`"

    def get_default_header(self, page: mknodes.MkPage):
        return ""

    def check_if_apply(self, page: mknodes.MkPage):
        return bool(classhelpers.get_qt_parent_class(self.item))


class QtScreenshotContainerProcessor(processors.ContainerProcessor):
    ID = "qt_signature"

    def append_block(self, page: mknodes.MkPage):
        if widget := self.item.setup_example():
            page += mknodes.MkCode.for_object(self.item.setup_example)
            page += prettyqtmarkdown.MkWidgetScreenShot(widget)

    def get_default_header(self, page: mknodes.MkPage):
        return "🖼 Screenshot"

    def check_if_apply(self, page: mknodes.MkPage):
        return (
            "setup_example" in self.item.__dict__
            and "Abstract" not in self.item.__name__
            and not self.item.__name__.endswith("Mixin")
        )


class QtContainerProcessor(processors.ContainerProcessor):
    ID = "qt"

    def append_block(self, page: mknodes.MkPage):
        if issubclass(self.item, itemmodels.SliceIdentityProxyModel):
            page += mknodes.MkAdmonition(SLICE_PROXY_INFO)
        if issubclass(self.item, core.AbstractItemModelMixin) and self.item.IS_RECURSIVE:
            page += mknodes.MkAdmonition(RECURSIVE_MODEL_INFO, typ="warning")
        if (
            issubclass(self.item, core.AbstractItemModelMixin)
            and self.item.DELEGATE_DEFAULT is not None
        ):
            msg = f"Recommended delegate: {self.item.DELEGATE_DEFAULT!r}"
            page += mknodes.MkAdmonition(msg)
        if issubclass(self.item, core.AbstractItemModelMixin) and hasattr(
            self.item, "SUPPORTS"
        ):
            page += mknodes.MkAdmonition(f"Supported data type: `{self.item.SUPPORTS}`")
        if issubclass(self.item, core.QObject):
            header = "⌗ Property table"
            table = prettyqtmarkdown.MkPropertyTable(self.item, header=header)
            page += table
        if hasattr(self.item, "ID") and issubclass(self.item, gui.Validator):
            page += f"\n\nValidator ID: **{self.item.ID}**\n\n"
        if hasattr(self.item, "ID") and issubclass(
            self.item, widgets.AbstractItemDelegateMixin
        ):
            page += f"\n\nDelegate ID: **{self.item.ID}**\n\n"

    def get_default_header(self, page: mknodes.MkPage):
        return ""


class MkPrettyQtClassPage(mknodes.MkClassPage):
    """A ClassPage specifically for Qt-based classes."""

    def get_pageprocessors(self):
        processors = super().get_pageprocessors()
        return [
            QtParentContainerProcessor(self.klass),
            *processors,
            QtContainerProcessor(self.klass),
            QtScreenshotContainerProcessor(self.klass),
        ]


if __name__ == "__main__":
    page = MkPrettyQtClassPage(klass=core.StringListModel)
    print(page.to_markdown())



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import datatypes


class GraphicsScale(widgets.GraphicsTransformMixin, widgets.QGraphicsScale):
    def set_origin(self, origin: datatypes.Vector3DType):
        self.setOrigin(datatypes.to_vector3d(origin))


if __name__ == "__main__":
    scale = GraphicsScale()
    scale.set_origin((1, 1, 1))



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from collections.abc import Iterator


FieldAlignmentStr = Literal["left", "right", "center", "accounting_style"]

FIELD_ALIGNMENT: bidict[FieldAlignmentStr, QtCore.QTextStream.FieldAlignment] = bidict(
    left=QtCore.QTextStream.FieldAlignment.AlignLeft,
    right=QtCore.QTextStream.FieldAlignment.AlignRight,
    center=QtCore.QTextStream.FieldAlignment.AlignCenter,
    accounting_style=QtCore.QTextStream.FieldAlignment.AlignAccountingStyle,
)

NumberFlagStr = Literal[
    "show_base", "force_point", "force_sign", "uppercase_base", "uppercase_digits"
]

NUMBER_FLAGS: bidict[NumberFlagStr, QtCore.QTextStream.NumberFlag] = bidict(
    show_base=QtCore.QTextStream.NumberFlag.ShowBase,
    force_point=QtCore.QTextStream.NumberFlag.ForcePoint,
    force_sign=QtCore.QTextStream.NumberFlag.ForceSign,
    uppercase_base=QtCore.QTextStream.NumberFlag.UppercaseBase,
    uppercase_digits=QtCore.QTextStream.NumberFlag.UppercaseDigits,
)

RealNumberNotationStr = Literal["scientific", "fixed", "smart"]

REAL_NUMBER_NOTATION: bidict[
    RealNumberNotationStr, QtCore.QTextStream.RealNumberNotation
] = bidict(
    scientific=QtCore.QTextStream.RealNumberNotation.ScientificNotation,
    fixed=QtCore.QTextStream.RealNumberNotation.FixedNotation,
    smart=QtCore.QTextStream.RealNumberNotation.SmartNotation,
)

StatusStr = Literal["ok", "read_past_end", "read_corrupt_data", "write_failed"]

STATUS: bidict[StatusStr, QtCore.QTextStream.Status] = bidict(
    ok=QtCore.QTextStream.Status.Ok,
    read_past_end=QtCore.QTextStream.Status.ReadPastEnd,
    read_corrupt_data=QtCore.QTextStream.Status.ReadCorruptData,
    write_failed=QtCore.QTextStream.Status.WriteFailed,
)


class TextStream(QtCore.QTextStream):
    """Convenient interface for reading and writing text."""

    def set_field_alignment(
        self, alignment: FieldAlignmentStr | QtCore.QTextStream.FieldAlignment
    ):
        """Set the field alignment.

        Args:
            alignment: field alignment
        """
        self.setFieldAlignment(FIELD_ALIGNMENT.get_enum_value(alignment))

    def get_field_alignment(self) -> FieldAlignmentStr:
        """Get current field alignment.

        Returns:
            current field alignment
        """
        return FIELD_ALIGNMENT.inverse[self.fieldAlignment()]

    def set_status(self, status: StatusStr | QtCore.QTextStream.Status):
        """Set the status.

        Args:
            status: status
        """
        self.setStatus(STATUS.get_enum_value(status))

    def get_status(self) -> StatusStr:
        """Get current status.

        Returns:
            current status
        """
        return STATUS.inverse[self.status()]

    def set_real_number_notation(
        self, notation: RealNumberNotationStr | QtCore.QTextStream.RealNumberNotation
    ):
        """Set the real number notation.

        Args:
            notation: real number notation
        """
        self.setRealNumberNotation(REAL_NUMBER_NOTATION.get_enum_value(notation))

    def get_real_number_notation(self) -> RealNumberNotationStr:
        """Get current real number notation.

        Returns:
            current real number notation
        """
        return REAL_NUMBER_NOTATION.inverse[self.realNumberNotation()]

    def read_lines(self) -> Iterator[str]:
        while True:
            if msg := self.readLine():
                yield msg
            else:
                return

    def get_number_flags(self) -> list[NumberFlagStr]:
        return NUMBER_FLAGS.get_list(self.numberFlags())

    def set_number_flags(self, *flags: NumberFlagStr):
        flag = NUMBER_FLAGS.merge_flags(flags)
        self.setNumberFlags(flag)


if __name__ == "__main__":
    stream = TextStream(QtCore.QByteArray(b"Test"))
    print(repr(stream))



"""Provides QtAsyncio classes and functions."""

from prettyqt.qt import PYSIDE6


if PYSIDE6:
    from PySide6.QtAsyncio import *  # noqa: F403  # type: ignore
else:
    msg = "No Qt bindings could be found"
    raise ModuleNotFoundError(msg)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from prettyqt import constants, core, iconprovider, itemmodels, widgets
from prettyqt.eventfilters import listviewgridresizeeventfilter
from prettyqt.utils import colors


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


# TODO: Set icon colour and copy code with color kwarg

ALL_COLLECTIONS = "All"


class IconBrowser(widgets.MainWindow):
    """A small browser window that allows the user to search through all icons.

    You can also copy the name and python code for the currently selected icon.
    """

    def __init__(self):
        super().__init__()
        self.setMinimumSize(500, 500)
        self.set_title("Icon Browser")
        font_maps = {k: v.charmap for k, v in iconprovider._instance().fonts.items()}
        icon_names = [
            f"{font_collection}.{icon_name}"
            for font_collection, font_data in font_maps.items()
            for icon_name in font_data
        ]
        model = IconModel(icon_color="text_role")
        model.setStringList(sorted(icon_names))

        self._proxy_model = itemmodels.FuzzyFilterProxyModel()
        self._proxy_model.setSourceModel(model)
        self._proxy_model.set_filter_case_sensitive(True)
        self._proxy_model.set_match_color(None)

        self._listview = widgets.ListView(
            self,
            uniform_item_sizes=True,
            view_mode="icon",
            context_menu_policy="custom",
            double_clicked=self._copy_icon_text,
            vertical_scroll_bar_policy="always_on",
        )
        eventfilter = listviewgridresizeeventfilter.ListViewGridResizeEventFilter(
            parent=self._listview
        )
        self._listview.installEventFilter(eventfilter)
        self._listview.set_model(self._proxy_model)

        self._lineedit = widgets.LineEdit(
            parent=self, text_changed=self._trigger_instant_update
        )

        self._combobox = widgets.ComboBox(
            parent=self,
            minimum_width=75,
            current_index_changed=self._trigger_instant_update,
        )
        self._combobox.addItems([ALL_COLLECTIONS, *sorted(font_maps.keys())])

        search_bar_frame = widgets.Frame(self)
        layout = search_bar_frame.set_layout("horizontal", margin=0)
        layout.add(self._combobox)
        layout.add(self._lineedit)

        self._copy_button = widgets.PushButton("Copy Name", clicked=self._copy_icon_text)
        frame = widgets.Frame(self)
        layout = frame.set_layout("vertical")
        layout.add(search_bar_frame)
        layout.add(self._listview)
        layout.add(self._copy_button)
        self.setCentralWidget(frame)
        self.add_shortcut("return", callback=self._copy_icon_text)
        self._lineedit.setFocus()
        self.position_on("screen")

    def _trigger_instant_update(self):
        """Stop timer used for committing search term and update proxy model instantly."""
        self._proxy_model.set_search_term(self._lineedit.text())

    def _copy_icon_text(self):
        """Copy the name of the currently selected icon to the clipboard."""
        if indexes := self._listview.selectedIndexes():
            widgets.Application.copy_to_clipboard(indexes[0].data())


class IconModel(core.StringListModel):
    def __init__(self, icon_color: datatypes.ColorType):
        super().__init__()
        self._icon_color = colors.get_color(icon_color).as_qt()

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        return constants.IS_ENABLED | constants.IS_SELECTABLE  # type: ignore

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role == constants.DECORATION_ROLE:
            icon_string = self.data(index, role=constants.DISPLAY_ROLE)
            return iconprovider._icon(icon_string, color=self._icon_color)
        return super().data(index, role)


def run():
    app = widgets.app()
    browser = IconBrowser()
    browser.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    run()



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Any

from prettyqt import core, gui, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class FileChooserButton(widgets.Widget):
    value_changed = core.Signal(pathlib.Path)

    def __init__(
        self,
        extensions: dict[str, list[str]] | None = None,
        mode: widgets.filedialog.AcceptModeStr = "save",
        file_mode: widgets.filedialog.FileModeStr = "existing_files",
        root: None | str | pathlib.Path = None,
        object_name: str = "file_chooser_button",
        **kwargs: Any,
    ):
        """Initialize FileChooserButton.

        Args:
            extensions: dict allowed extensions
                        form: "'name': ['.ext1', '.ext2']"
            mode: Accept mode ("save" or "load")
            file_mode: File mode ("existing_files", "existing_file", "any_file",
                                  or "directory")
            root: Root path
            object_name: Object name
            kwargs: kwargs passed to parent
        """
        super().__init__(object_name=object_name, **kwargs)
        self._path: pathlib.Path | None = None
        self._extensions = extensions or []
        self.mode = mode
        self.file_mode = file_mode
        self.root = root
        layout = widgets.HBoxLayout(self)
        layout.set_margin(0)
        self.lineedit = widgets.LineEdit()
        self.lineedit.set_read_only()
        layout.add(self.lineedit)
        action = gui.Action(triggered=self.open_file)
        if self.file_mode == "directory":
            action.set_icon("mdi.folder-outline")
        else:
            action.set_icon("mdi.file-outline")

        self.button = widgets.ToolButton()
        self.button.setDefaultAction(action)
        layout.add(self.button)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, pathlib.Path)

    @core.Slot()
    def open_file(self):
        dialog = widgets.FileDialog(
            parent=self,
            path_id="file_path",
            mode=self.mode,
            path=self.root,
            file_mode=self.file_mode,
        )
        if self._extensions:
            dialog.set_extension_filter(self._extensions)
        if not dialog.choose():
            return
        self.set_path(dialog.selected_file())
        self.value_changed.emit(self._path)

    def set_path(self, path: datatypes.PathType | None):
        if path is None:
            self._path = None
            self.lineedit.set_text("")
        else:
            self._path = pathlib.Path(path)
            self.lineedit.set_text(os.fspath(path))

    def get_value(self) -> pathlib.Path | None:
        return self._path

    def set_value(self, value: datatypes.PathType | None):
        self.set_path(value)

    def get_extensions(self) -> list[str]:
        return self._extensions

    def set_extensions(self, extensions: list[str]):
        self._extensions = extensions

    # path = core.Property(object, get_value, set_value)
    # extensions = core.Property(object, get_extensions, set_extensions)


if __name__ == "__main__":
    app = widgets.app()
    btn = FileChooserButton()
    btn.show()
    btn.value_changed.connect(print)
    app.exec()



from __future__ import annotations

from typing import Self

from prettyqt import constants
from prettyqt.qt import QtCore


class Date(QtCore.QDate):
    """Date functions."""

    def __repr__(self):
        template = super().__repr__().split("(")[1]  # type: ignore
        return f"{type(self).__name__}({template}"

    def __str__(self):
        return self.toString("yyyy-MM-dd")

    def __format__(self, format_spec: str):
        return self.toString(format_spec)

    @property
    def _year(self) -> int:
        return self.year()

    @property
    def _month(self) -> int:
        return self.month()

    @property
    def _day(self) -> int:
        return self.day()

    __match_args__ = ("_year", "_month", "_day")

    def __reduce__(self):
        return type(self), (self.year(), self.month(), self.day())

    def add_days(self, days: int) -> Self:
        return type(self)(self.addDays(days))

    def add_months(self, months: int, calendar: QtCore.QCalendar | None = None) -> Self:
        if calendar:
            return type(self)(self.addMonths(months, calendar))
        return type(self)(self.addMonths(months))

    def add_years(self, years: int, calendar: QtCore.QCalendar | None = None) -> Self:
        if calendar:
            return type(self)(self.addYears(years, calendar))
        return type(self)(self.addYears(years))

    @classmethod
    def get_current_date(cls) -> Self:
        return cls(cls.currentDate())

    @classmethod
    def from_string(cls, text: str, date_format: constants.DateFormatStr | str) -> Self:
        if date_format in constants.DATE_FORMAT:
            date_format = constants.DATE_FORMAT[date_format]
        return cls(cls.fromString(text, date_format))

    def replace(
        self, year: int | None = None, month: int | None = None, day: int | None = None
    ):
        self.setDate(
            year if year is not None else self.year(),
            month if month is not None else self.month(),
            day if day is not None else self.day(),
        )


if __name__ == "__main__":
    dt = Date(2000, 11, 11)
    dt.add_years(30)



from __future__ import annotations

from prettyqt import custom_widgets, widgets


class StandardIconsWidget(widgets.Widget):
    """Dialog showing standard icons."""

    def __init__(self, parent: widgets.QWidget | None = None):
        super().__init__(parent)
        layout = custom_widgets.MultiLineLayout(parent=self)
        for k in widgets.style.STANDARD_PIXMAP:
            icon_layout = widgets.HBoxLayout()
            icon = widgets.Application.get_style_icon(k)
            label = widgets.Label(pixmap=icon.pixmap(32, 32))
            icon_layout.addWidget(label)
            icon_layout.addWidget(widgets.LineEdit(k))
            layout.addLayout(icon_layout)
        self.set_layout(layout)
        self.set_title("Standard Platform Icons")
        icon = widgets.Application.get_style_icon("titlebar_menu_button")
        self.set_icon(icon)


if __name__ == "__main__":
    app = widgets.app()
    widget = StandardIconsWidget()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQml
from prettyqt.utils import get_repr


class QmlError(QtQml.QQmlError):
    """Encapsulates a QML error."""

    def __repr__(self):
        return get_repr(self, self.toString())

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def get_property(self) -> core.MetaProperty:
        return core.MetaProperty(self.property())

    def __str__(self):
        return self.toString()


if __name__ == "__main__":
    error = QmlError()



from __future__ import annotations

from prettyqt import gui


class TextTableCell(gui.QTextTableCell):
    def get_format(self) -> gui.TextCharFormat:
        #  .format() seems to crash both bindings?
        return gui.TextCharFormat(self.format())

    def get_first_cursor_position(self) -> gui.TextCursor:
        return gui.TextCursor(self.firstCursorPosition())

    def get_last_cursor_position(self) -> gui.TextCursor:
        return gui.TextCursor(self.lastCursorPosition())


if __name__ == "__main__":
    app = gui.app()
    cell = TextTableCell()
    cell.get_format()



from __future__ import annotations

from importlib import resources
import pathlib


ROOT_PATH = pathlib.Path(resources.files("prettyqt"))  # type: ignore

LOCALIZATION_PATH = ROOT_PATH / "localization"
THEMES_PATH = ROOT_PATH / "resources" / "themes"
RE_LEXER_PATH = (
    ROOT_PATH / "syntaxhighlighters" / "custom_highlighters" / "regularexpressionlexer.py"
)
ICON_FONT_PATH = ROOT_PATH / "iconprovider" / "fonts"

DOCS_PATH = ROOT_PATH.parent / "docs"
INV_FILE = DOCS_PATH / "qt6.inv"



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import iconprovider, widgets
from prettyqt.qt import QtCore


if TYPE_CHECKING:
    from prettyqt import gui


class AwesomeFileIconProvider(widgets.FileIconProvider):
    def __init__(self):
        super().__init__()
        self.folder_icon = iconprovider.get_icon("mdi.folder")
        self.text_icon = iconprovider.get_icon("mdi.file-document")
        self.file_icon = iconprovider.get_icon("mdi.file")
        self.desktop_icon = iconprovider.get_icon("mdi.desktop-mac-dashboard")
        self.computer_icon = iconprovider.get_icon("mdi.desktop-classic")
        self.trashcan_icon = iconprovider.get_icon("mdi.trash-can")
        self.drive_icon = iconprovider.get_icon("mdi.harddisk")
        self.network_icon = iconprovider.get_icon("mdi.folder-network")

    def icon(  # noqa: PLR0911
        self, icon: widgets.QFileIconProvider.IconType | QtCore.QFileInfo
    ) -> gui.Icon:
        match icon:
            case widgets.FileIconProvider.IconType.Folder:
                return self.folder_icon
            case widgets.FileIconProvider.IconType.File:
                return self.file_icon
            case widgets.FileIconProvider.IconType.Computer:
                return self.computer_icon
            case widgets.FileIconProvider.IconType.Desktop:
                return self.desktop_icon
            case widgets.FileIconProvider.IconType.Trashcan:
                return self.trashcan_icon
            case widgets.FileIconProvider.IconType.Network:
                return self.network_icon
            case widgets.FileIconProvider.IconType.Drive:
                return self.drive_icon
            case QtCore.QFileInfo():
                if icon.isDir():
                    return self.folder_icon
                if icon.isFile():
                    return self.file_icon
                return self.file_icon
            case _:
                return self.file_icon


if __name__ == "__main__":
    from prettyqt import custom_widgets, widgets

    app = widgets.app()
    app.load_language("de")
    model = widgets.FileSystemModel()
    model.set_root_path("root")
    prov = AwesomeFileIconProvider()
    model.setIconProvider(prov)
    tree = custom_widgets.FileTree()
    tree.set_model(model)
    tree.show()
    app.exec()



from __future__ import annotations

import importlib.util
import logging

from qtconsole.inprocess import QtInProcessKernelManager

from prettyqt import core, ipython, widgets


# import asyncio


logger = logging.getLogger(__name__)


class InProcessIPythonWidget(ipython.BaseIPythonWidget):
    """Convenience class for a live IPython console widget."""

    evaluated = core.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel(show_banner=False)

        self.kernel_manager.kernel.log.setLevel(logging.CRITICAL)
        self.kernel_manager.kernel.gui = "qt"
        if importlib.util.find_spec("matplotlib"):
            self.kernel_manager.kernel.shell.enable_matplotlib(gui="inline")
        self.kernel_client = self.kernel_manager.client()
        self.kernel_client.start_channels()

    def shutdown(self):
        logger.info("shutting down IPython kernel...")
        assert self.kernel_manager
        assert self.kernel_client
        self.kernel_client.stop_channels()
        self.kernel_manager.shutdown_kernel()
        logger.info("shutdown successful.")

    def push_vars(self, var_dict):
        """Send python objects to IPYthon namespace.

        Given a dictionary containing name / value pairs, push those variables
        to the IPython console widget.
        """
        assert self.kernel_manager
        self.kernel_manager.kernel.shell.push(var_dict)
        for key in var_dict:
            self._append_plain_text(f'\nadded "{key}" object to namespace\n', True)

    def eval(self, obj_name):
        """Pull object with name *obj_name from namespace."""
        assert self.kernel_manager
        try:
            data = self.kernel_manager.kernel.shell.ev(obj_name)
        except NameError:
            logger.exception("Exception when evaluating %s", obj_name)
            return None
        self._append_plain_text(f'\nread "{obj_name}" object to namespace\n', True)
        self.evaluated.emit(data)
        return data


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("fusion")
    console_widget = InProcessIPythonWidget(app)
    console_widget.print_text("hallo")
    console_widget.show()
    console_widget.evaluated.connect(print)
    console_widget.completion_mode = "ncurses"
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import time

from prettyqt import core


class DebouncedSignal:
    """Signal definition.

    Instances of this class will be replaced with a QObject holding a SignalInstance
     by the SignalMeta metaclass.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class SignalMeta(type(core.QObject)):
    def __new__(cls, name, bases, attrs, **kwargs):
        to_replace = {
            k: (v.args, v.kwargs)
            for k, v in attrs.items()
            if isinstance(v, DebouncedSignal)
        }
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        user_init = new_cls.__init__

        def __init__(self, *args, **kwargs_):  # noqa: N807
            user_init(self, *args, **kwargs_)
            for k, v in to_replace.items():
                self.__dict__[k] = get_debounced_signal(*v[0], **v[1])()

        new_cls.__init__ = __init__
        return new_cls


class BaseSignalImpl(core.Object):
    def __getitem__(self, item: type | str):
        return self.signal[item]

    def __getattr__(self, attr):
        return getattr(self.signal, attr)

    def connect(self, slot_or_signal):
        if isinstance(slot_or_signal, BaseSignalImpl):
            self.signal.connect(slot_or_signal.signal)  # or .emit?
        else:
            self.signal.connect(slot_or_signal)

    # def __call__(self, *args, **kwargs):
    #     self.emit(*args)
    #     print("__call__", args)


def get_debounced_signal(
    *args: type | str,
    debounce_time: int | None = None,
    repeat_same: bool = True,
    **kwargs,
):
    class DebouncedSignalImpl(BaseSignalImpl):
        signal = core.Signal(*args)

        def __init__(self, *args):
            super().__init__()
            self.last_tried_emission = ...  # just a sentinel value
            self.debounce_time = debounce_time
            # self.last_emission_time = time.time()
            # self.last_emission_try = time.time()
            self.timer = core.Timer(timeout=self.timer_emit, single_shot=True)
            self.repeat_same = repeat_same

        def emit(self, *args):
            # now = time.time()
            # debounce_time = self.debounce_time or 0
            # should_emit = now - self.last_emission_time > debounce_time
            if (args != self.last_tried_emission) or self.repeat_same:
                if self.debounce_time:
                    self.timer.start(self.debounce_time)
                else:
                    self.signal.emit(*args)
            self.last_tried_emission = args
            # print("emitted", args)
            # self.last_emission_time = time.time()

        def timer_emit(self):
            self.signal.emit(*self.last_tried_emission)
            self.last_emission_time = time.time()
            # print("emitted delayed", self.last_tried_emission)

    return DebouncedSignalImpl


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    textedit = widgets.PlainTextEdit()

    class EmitPushButton(widgets.PushButton, metaclass=SignalMeta):
        btn_clicked_delayed = DebouncedSignal(str, debounce_time=1000, repeat_same=False)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.clicked.connect(self.on_click)

        def on_click(self):
            # self.btn_clicked.emit(self.text())
            self.btn_clicked_delayed.emit(str(self.text()))

    # print(SignalInstance.test)
    btn1 = EmitPushButton("Signal 1")
    btn2 = EmitPushButton("Signal 2")
    widget = widgets.Widget()
    widget.set_layout("vertical")
    widget.box.add(textedit)
    widget.box.add(btn1)
    widget.box.add(btn2)
    # textedit.value_changed.connect(btn2.btn_clicked_delayed.emit)
    btn1.btn_clicked_delayed.connect(textedit.appendPlainText)
    btn2.btn_clicked_delayed.connect(textedit.appendPlainText)
    widget.show()
    # a.number.signal.emit("test")
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import functools
from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


DeletionPolicyStr = Literal["keep", "delete"]

DELETION_POLICY: bidict[DeletionPolicyStr, core.QAbstractAnimation.DeletionPolicy] = (
    bidict(
        keep=core.QAbstractAnimation.DeletionPolicy.KeepWhenStopped,
        delete=core.QAbstractAnimation.DeletionPolicy.DeleteWhenStopped,
    )
)

DirectionStr = Literal["forward", "backward"]

DIRECTION: bidict[DirectionStr, core.QAbstractAnimation.Direction] = bidict(
    forward=core.QAbstractAnimation.Direction.Forward,
    backward=core.QAbstractAnimation.Direction.Backward,
)

StateStr = Literal["stopped", "paused", "running"]

STATE: bidict[StateStr, core.QAbstractAnimation.State] = bidict(
    stopped=core.QAbstractAnimation.State.Stopped,
    paused=core.QAbstractAnimation.State.Paused,
    running=core.QAbstractAnimation.State.Running,
)


class AbstractAnimationMixin(core.ObjectMixin):
    """The base of all animations."""

    def __len__(self):
        return self.duration()

    def __and__(self, other: core.QAbstractAnimation) -> core.SequentialAnimationGroup:
        group = core.SequentialAnimationGroup()
        group.addAnimation(self)
        group.addAnimation(other)
        return group

    def __or__(self, other: core.QAbstractAnimation) -> core.ParallelAnimationGroup:
        group = core.ParallelAnimationGroup()
        group.addAnimation(self)
        group.addAnimation(other)
        return group

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"direction": DIRECTION, "state": STATE}
        return maps

    def toggle_direction(self):
        direction = AbstractAnimation.Direction
        is_forward = self.direction() == direction.Forward
        direction = direction.Backward if is_forward else direction.Forward
        self.setDirection(direction)

    def set_direction(self, direction: DirectionStr | core.QAbstractAnimation.Direction):
        """Set animation direction.

        Args:
            direction: animation direction
        """
        self.setDirection(DIRECTION.get_enum_value(direction))

    def get_direction(self) -> DirectionStr:
        """Get the current animation direction.

        Returns:
            animation direction
        """
        return DIRECTION.inverse[self.direction()]

    def get_state(self) -> StateStr:
        """Get the current animation state.

        Returns:
            animation state
        """
        return STATE.inverse[self.state()]

    def start_animation(
        self,
        policy: DeletionPolicyStr | core.QAbstractAnimation.DeletionPolicy = "keep",
        interval: int = 0,
        single_shot: bool = True,
    ):
        """Start the animation.

        Args:
            policy: animation policy
            interval: time interval / delay for timer
            single_shot: whether animation gets triggered once or in intervals
        """
        if policy in DELETION_POLICY:
            policy = DELETION_POLICY[policy]
        if interval:
            fn = functools.partial(self.start, policy)
            self.start_callback_timer(fn, interval, single_shot=single_shot)
        else:
            self.start(policy)

    def restart_animation(
        self,
        policy: DeletionPolicyStr | core.QAbstractAnimation.DeletionPolicy = "keep",
    ):
        """Restart the animation.

        Args:
            policy: animation policy
        """
        self.stop()
        self.start_animation(policy)

    def run(self, delay: int = 0, single_shot: bool = True):
        self.start_callback_timer(self.start, delay, single_shot=single_shot)


class AbstractAnimation(AbstractAnimationMixin, core.QAbstractAnimation):
    pass


if __name__ == "__main__":
    anim1 = core.PropertyAnimation()
    anim2 = core.PropertyAnimation()
    group = anim1 | anim2
    print(type(group))



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.utils import helpers


class Timer(core.ObjectMixin, core.QTimer):
    """Repetitive and single-shot timers."""

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"timerType": constants.TIMER_TYPE}
        return maps

    def set_type(self, typ: constants.TimerTypeStr | constants.TimerType):
        """Set the timer type.

        Args:
            typ: timer type
        """
        self.setTimerType(constants.TIMER_TYPE.get_enum_value(typ))

    def get_type(self) -> constants.TimerTypeStr:
        """Return current timer type.

        Returns:
            timer type
        """
        return constants.TIMER_TYPE.inverse[self.timerType()]

    def set_interval(self, interval: int | str):
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        self.setInterval(interval)

    def start_timer(self, interval: None | int | str = None):
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        if interval is None:
            self.start()
        else:
            self.start(interval)

    def restart(self):
        self.stop()
        self.start()


if __name__ == "__main__":
    timer = Timer(timeout=print)
    timer.set_interval("2m")



from __future__ import annotations

from prettyqt import constants, core, widgets
from prettyqt.utils import listdelegators


class TreeWidgetMixin(widgets.TreeViewMixin):
    def __contains__(self, other: widgets.QTreeWidgetItem):
        return self.indexOfTopLevelItem(other) >= 0

    def __getitem__(
        self, index: int | slice
    ) -> widgets.QTreeWidgetItem | listdelegators.ListDelegator[widgets.QTreeWidgetItem]:
        match index:
            case int():
                item = self.topLevelItem(index)
                if item is None:
                    raise KeyError(index)
                return item
            case slice():
                count = self.topLevelItemCount() if index.stop is None else index.stop
                values = list(range(count)[index])
                ls = [self.topLevelItem(i) for i in values]
                return listdelegators.ListDelegator(ls)
            case _:
                raise TypeError(index)

    def sort(self, column: int = 0, reverse: bool = False):
        order = constants.DESCENDING if reverse else constants.ASCENDING
        self.sortItems(column, order)

    def find_items(
        self,
        text: str,
        column: int = 0,
        mode: constants.MatchFlagStr | constants.MatchFlag = "exact",
        recursive: bool = False,
        case_sensitive: bool = False,
    ) -> listdelegators.ListDelegator[widgets.QTreeWidgetItem]:
        flag = constants.MATCH_FLAGS.get_enum_value(mode)
        if recursive:
            flag |= constants.MatchFlag.MatchRecursive
        if case_sensitive:
            flag |= constants.MatchFlag.MatchCaseSensitive
        items = self.findItems(text, flag, column)
        return listdelegators.ListDelegator(items)

    def get_items(
        self, recursive: bool = False
    ) -> listdelegators.ListDelegator[widgets.QTreeWidgetItem]:
        """Get TreeWidgetItems of this widget.

        Arguments:
            recursive: whether to include all items of the tree.

        recursive option is written iteratively to also support original QTreeWidgetItems.
        """
        root = self.invisibleRootItem(())
        if not recursive:
            return [root.child(i) for i in range(root.childCount())]
        results = []
        nodes = [root]
        while nodes:
            items = []
            for node in nodes:
                results.append(node)
                items.extend(node.child(i) for i in range(node.childCount()))
            nodes = items
        return listdelegators.ListDelegator(results[1:])

    def scroll_to_item(
        self,
        item: widgets.QTreeWidgetItem,
        scroll_hint: widgets.abstractitemview.ScrollHintStr
        | widgets.QAbstractItemView.ScrollHint = "ensure_visible",
    ):
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT[scroll_hint])

    def removeTopLevelItem(self, item):
        for i in range(self.topLevelItemCount()):
            if self.topLevelItem(i) is item:
                self.takeTopLevelItem(i)
                return
        msg = f"Item {item!r} not in top-level items."
        raise IndexError(msg)

    def openPersistentEditor(
        self, index: core.ModelIndex | widgets.QTreeWidgetItem, column: int = 0
    ):
        if isinstance(index, core.ModelIndex):
            column = index.column()
            index = self.itemFromIndex(index)
        super().openPersistentEditor(index, column)

    def closePersistentEditor(
        self, index: core.ModelIndex | widgets.QTreeWidgetItem, column: int = 0
    ):
        if isinstance(index, core.ModelIndex):
            column = index.column()
            index = self.itemFromIndex(index)
        super().closePersistentEditor(index, column)

    def isPersistentEditorOpen(
        self, index: core.ModelIndex | widgets.QTreeWidgetItem, column: int = 0
    ) -> bool:
        if isinstance(index, core.ModelIndex):
            column = index.column()
            index = self.itemFromIndex(index)
        return super().isPersistentEditorOpen(index, column)


class TreeWidget(TreeWidgetMixin, widgets.QTreeWidget):
    """Tree view that uses a predefined tree model."""


if __name__ == "__main__":
    app = widgets.app()
    widget = TreeWidget()
    widget.openPersistentEditor(core.ModelIndex(), 1)
    widget.show()
    app.exec()



from __future__ import annotations

import html
import os
from typing import TYPE_CHECKING, Literal

from prettyqt import constants, core, gui, iconprovider
from prettyqt.utils import bidict, get_repr


if TYPE_CHECKING:
    from collections.abc import Callable

    from prettyqt.utils import datatypes


ActionEventStr = Literal["trigger", "hover"]

ACTION_EVENT: bidict[ActionEventStr, gui.QAction.ActionEvent] = bidict(
    trigger=gui.QAction.ActionEvent.Trigger,
    hover=gui.QAction.ActionEvent.Hover,
)

PriorityStr = Literal["low", "normal", "high"]

PRIORITIES: bidict[PriorityStr, gui.QAction.Priority] = bidict(
    low=gui.QAction.Priority.LowPriority,
    normal=gui.QAction.Priority.NormalPriority,
    high=gui.QAction.Priority.HighPriority,
)

MenuRoleStr = Literal[
    "none",
    "text_heuristic",
    "application_specific",
    "about_qt",
    "about",
    "preferences",
    "quit",
]

MENU_ROLE: bidict[MenuRoleStr, gui.QAction.MenuRole] = bidict(
    none=gui.QAction.MenuRole.NoRole,
    text_heuristic=gui.QAction.MenuRole.TextHeuristicRole,
    application_specific=gui.QAction.MenuRole.ApplicationSpecificRole,
    about_qt=gui.QAction.MenuRole.AboutQtRole,
    about=gui.QAction.MenuRole.AboutRole,
    preferences=gui.QAction.MenuRole.PreferencesRole,
    quit=gui.QAction.MenuRole.QuitRole,
)


class ActionMixin(core.ObjectMixin):
    def __init__(
        self,
        *args,
        text: str | None = None,
        icon: datatypes.IconType = None,
        callback: Callable | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if callback is not None:
            self.triggered.connect(callback)
        self._usage_count = 0
        if text:
            self.setText(text)
        if icon:
            self.set_icon(icon)
        self.triggered.connect(self._increase_usage_counter)

    def __repr__(self) -> str:
        return get_repr(self, self.text())

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "priority": PRIORITIES,
            "shortcutContext": constants.SHORTCUT_CONTEXT,
            "menuRole": MENU_ROLE,
        }
        return maps

    def _increase_usage_counter(self):
        self._usage_count += 1

    def get_usage_count(self) -> int:
        return self._usage_count

    def get_type(self) -> Literal["menu", "separator", "widget", "regular"]:
        if self.menu() is not None:
            return "menu"
        if self.isSeparator():
            return "separator"
        if hasattr(self, "defaultWidget"):
            return "widget"
        return "regular"

    def set_disabled(self):
        self.setEnabled(False)

    def set_enabled(self, enabled: bool = True):
        self.setEnabled(enabled)

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, core.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        tooltip = tooltip.replace("\n", "<br/>")
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    setToolTip = set_tooltip  # noqa: N815

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the action.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        super().setIcon(icon)

    setIcon = set_icon  # noqa: N815

    def set_shortcut(self, shortcut: None | gui.QKeySequence | str):
        if shortcut is None:
            shortcut = ""
        if isinstance(shortcut, str):
            shortcut = gui.KeySequence(
                shortcut, gui.KeySequence.SequenceFormat.PortableText
            )
        super().setShortcut(shortcut)

    setShortcut = set_shortcut  # noqa: N815

    def setText(self, text: str | None):
        super().setText(text or "")

    set_text = setText

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def get_shortcut(self) -> gui.KeySequence | None:
        shortcut = self.shortcut()
        return (
            gui.KeySequence(
                shortcut.toString(), gui.KeySequence.SequenceFormat.PortableText
            )
            if shortcut
            else None
        )

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def set_menu(self, menu):
        self.setMenu(menu)

    def set_priority(self, priority: PriorityStr | gui.QAction.Priority):
        """Set priority of the action.

        Args:
            priority: priority for the action
        """
        super().setPriority(PRIORITIES.get_enum_value(priority))

    setPriority = set_priority  # noqa: N815

    def get_priority(self) -> PriorityStr:
        """Return current priority.

        Returns:
            priority
        """
        return PRIORITIES.inverse[self.priority()]

    def set_shortcut_context(
        self, context: constants.ShortcutContextStr | constants.ShortcutContext
    ):
        """Set shortcut context.

        Args:
            context: shortcut context
        """
        super().setShortcutContext(constants.SHORTCUT_CONTEXT.get_enum_value(context))

    setShortcutContext = set_shortcut_context  # noqa: N815

    def get_shortcut_context(self) -> constants.ShortcutContextStr:
        """Return shortcut context.

        Returns:
            shortcut context
        """
        return constants.SHORTCUT_CONTEXT.inverse[super().shortcutContext()]

    def set_menu_role(self, role: MenuRoleStr | gui.QAction.MenuRole):
        """Set menu role.

        Args:
            role: menu role
        """
        super().setMenuRole(MENU_ROLE.get_enum_value(role))

    setMenuRole = set_menu_role  # noqa: N815

    def get_menu_role(self) -> MenuRoleStr:
        """Return menu role.

        Returns:
            menu role
        """
        return MENU_ROLE.inverse[super().menuRole()]

    def show_shortcut_in_contextmenu(self, state: bool = True):
        self.setShortcutVisibleInContextMenu(state)

    usage_count = core.Property(
        int,
        get_usage_count,
        doc="How often the action was triggered",
    )


class Action(ActionMixin, gui.QAction):
    """Abstraction for user commands that can be added to different GUI components."""


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    action = Action(text="This is a test", shortcut=None)



"""PrettyQt Validators.

PrettyQt ships a large number of predefined validators.

Apart from LineEdits, ComboBoxes and SpinBoxes, PlainTextEdits also gained the ability
to take a validator.

The set_validator method also gained a "strict" keyword argument, which wraps the passed
validator into a NotStrictValidator, effectively removing the "Invalid" state. That way
there never is a situation where the user input is ignored, which might be preferred in a
lot of situations.

The acceptance of "" may also be overriden explicitely by setting the "allow_empty"
keyword argument.

Validators can also be combined. The resulting CompositeValidator checks if all
containing validators accept the input.


Example:
```py
val_1 = BlackListValidator(["Blacklisted word"])
val_2 = AlphanumericValidator()
and_validator = val_1 & val_2  # returns an AndValidator

val_1 = AlphanumericValidator()
val_2 = RegularExpressionValidator("some_regex")
or_val = val_1 | val_2 # returns an OrValidator
```
Validators can also be set by an id. Since everything is typed with Literals,
the possible ids should be shown by your IDE.
```py
lineedit.set_validator("float")
```

"""

from __future__ import annotations

from .alphanumericvalidator import AlphaNumericValidator
from .blacklistvalidator import BlacklistValidator, NotEmptyValidator, NotZeroValidator
from .colorvalidator import ColorValidator
from .compositevalidator import AndValidator, OrValidator
from .functionvalidator import FunctionValidator
from .hexvalidator import HexValidator
from .integervalidator import IntegerValidator
from .jsonvalidator import JsonValidator
from .monotoniclistvalidator import MonotonicListValidator
from .notstrictvalidator import NotStrictValidator
from .pathvalidator import PathValidator
from .pythoncodevalidator import PythonCodeValidator
from .qssvalidator import QssValidator
from .regexpatternvalidator import RegexPatternValidator
from .regexvalidators import FloatListValidator, IntListValidator
from .scientificvalidators import ScientificFloatValidator, ScientificIntegerValidator
from .textlengthvalidator import TextLengthValidator
from .whitelistvalidator import EmptyValidator, WhitelistValidator


__all__ = [
    "AlphaNumericValidator",
    "AndValidator",
    "BlacklistValidator",
    "ColorValidator",
    "EmptyValidator",
    "FloatListValidator",
    "FunctionValidator",
    "HexValidator",
    "IntListValidator",
    "IntegerValidator",
    "JsonValidator",
    "MonotonicListValidator",
    "NotEmptyValidator",
    "NotStrictValidator",
    "NotZeroValidator",
    "OrValidator",
    "PathValidator",
    "PythonCodeValidator",
    "QssValidator",
    "RegexPatternValidator",
    "ScientificFloatValidator",
    "ScientificIntegerValidator",
    "TextLengthValidator",
    "WhitelistValidator",
]



from __future__ import annotations

import re

from prettyqt import gui


re_custom_sep = re.compile(r"\s*,\s*")


class MonotonicListValidator(gui.Validator):
    """Validator which checks whether given string contains a monotonic list."""

    ID = "monotonic"

    def __init__(self, kind: str = "increasing", parent=None):
        super().__init__(parent)
        self._kind = kind

    def __eq__(self, other: object):
        return isinstance(other, MonotonicListValidator) and other._kind == self._kind

    def validate(self, string: str, pos: int) -> tuple[gui.Validator.State, str, int]:
        if self._kind != "increasing":
            string = string[::-1]
        for i, c in enumerate(string, start=1):
            if c not in "+-., 0123456789":
                return self.State.Invalid, string, i
        if pos == len(string) >= 2 and string[-1] == " " and string[-2].isdigit():  # noqa: PLR2004
            string = f"{string[:-1]}, "
            pos += 1
        prev = None
        for valuestr in re_custom_sep.split(string.strip()):
            try:
                value = float(valuestr)
            except ValueError:
                return self.State.Intermediate, string, pos
            if prev is not None and value <= prev:
                return self.State.Intermediate, string, pos
            prev = value
        return self.State.Acceptable, string, pos


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    val = MonotonicListValidator(kind="decreasing")
    print(val.validate("1,0", 0))



from __future__ import annotations

from prettyqt import constants, core, widgets


class CheckBox(widgets.AbstractButtonMixin, widgets.QCheckBox):
    """Checkbox with a text label."""

    value_changed = core.Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stateChanged.connect(self.value_changed)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, bool)

    def set_checkstate(self, state: constants.CheckStateStr | constants.CheckState):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use
        """
        self.setCheckState(constants.CHECK_STATE.get_enum_value(state))

    def get_checkstate(self) -> constants.CheckStateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState()]


if __name__ == "__main__":
    app = widgets.app()
    widget = CheckBox("test")
    widget.show()
    app.exec()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from mknodes.utils import connector

from prettyqt import constants


if TYPE_CHECKING:
    from prettyqt import core


logger = logging.getLogger(__name__)


class IndexConnector(connector.Connector):
    def __init__(
        self,
        index: core.ModelIndex,
        fetch_more: bool = True,
        attribute_roles=None,
    ):
        self.fetch_more = fetch_more
        self.attribute_roles = attribute_roles or []
        super().__init__([index])

    def get_id(self, item: core.ModelIndex) -> int:
        return id(item)

    def get_title(self, item: core.ModelIndex) -> str:
        return item.data(constants.DISPLAY_ROLE)

    def get_attributes(self, item: core.ModelIndex) -> list[str]:
        return [item.data(role) for role in self.attribute_roles]

    def get_children(self, item: core.ModelIndex) -> list[core.ModelIndex]:
        model = item.model()
        if self.fetch_more:
            while model.canFetchMore(item):
                model.fetchMore(item)
        return [model.index(i, 0, item) for i in range(model.rowCount(item))]


class MkPrettyQtDiagram:
    pass


if __name__ == "__main__":
    import mknodes

    page = mknodes.MkPage()
    page += mknodes.MkAdmonition("test")



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtQuick
from prettyqt.utils import bidict


DIRTY_STATE_BIT = bidict(
    matrix=QtQuick.QSGNode.DirtyStateBit.DirtyMatrix,
    node_added=QtQuick.QSGNode.DirtyStateBit.DirtyNodeAdded,
    node_removed=QtQuick.QSGNode.DirtyStateBit.DirtyNodeRemoved,
    geometry=QtQuick.QSGNode.DirtyStateBit.DirtyGeometry,
    material=QtQuick.QSGNode.DirtyStateBit.DirtyMaterial,
    opacity=QtQuick.QSGNode.DirtyStateBit.DirtyOpacity,
    # subtree_blocked=QtQuick.QSGNode.DirtyStateBit.DirtySubtreeBlocked,
)

DirtyStateBitStr = Literal[
    "matrix",
    "node_added",
    "node_removed",
    "geometry",
    "material",
    "opacity",
    "subtree_blocked",
]

FLAG = bidict(
    owned_by_parent=QtQuick.QSGNode.Flag.OwnedByParent,
    use_preprocess=QtQuick.QSGNode.Flag.UsePreprocess,
    owns_geometry=QtQuick.QSGNode.Flag.OwnsGeometry,
    owns_material=QtQuick.QSGNode.Flag.OwnsMaterial,
    owns_opaque_material=QtQuick.QSGNode.Flag.OwnsOpaqueMaterial,
)

FlagStr = Literal[
    "owned_by_parent",
    "use_preprocess",
    "owns_geometry",
    "owns_material",
    "owns_opaque_material",
]

NODE_TYPE = bidict(
    basic=QtQuick.QSGNode.NodeType.BasicNodeType,
    geometry=QtQuick.QSGNode.NodeType.GeometryNodeType,
    transform=QtQuick.QSGNode.NodeType.TransformNodeType,
    clip=QtQuick.QSGNode.NodeType.ClipNodeType,
    opacity=QtQuick.QSGNode.NodeType.OpacityNodeType,
    # render=QtQuick.QSGNode.NodeType.RenderNodeType,
)

NodeTypeStr = Literal[
    "basic",
    "geometry",
    "transform",
    "clip",
    "opacity",
    "render",
]


class SGNode(QtQuick.QSGNode):
    """The base class for all nodes in the scene graph."""

    def __getitem__(self, index: int) -> QtQuick.QSGNode:
        return self.childAtIndex(index)

    def __delitem__(self, item: int | QtQuick.QSGNode):
        if isinstance(item, int):
            item = self.childAtIndex(item)
        self.removeChildNode(item)

    def get_type(self) -> NodeTypeStr:
        """Get the type of the node.

        Returns:
            Node type
        """
        return NODE_TYPE.inverse[self.type()]

    def get_flags(self) -> list[FlagStr]:
        return FLAG.get_list(self.flags())

    def get_children(self, recursive: bool = False) -> list[SGNode]:
        """Get children of this item.

        recursive option is written iteratively to also support original QTreeWidgetItems.
        """
        if not recursive:
            return [self.childAtIndex(i) for i in range(self.childCount())]
        results = []
        nodes = [self]
        while nodes:
            items = []
            for node in nodes:
                results.append(node)
                items.extend(node.childAtIndex(i) for i in range(node.childCount()))
            nodes = items
        return results[1:]


if __name__ == "__main__":
    node = SGNode()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class CodeEditor(widgets.PlainTextEdit):
    """Super basic code editor."""

    def __init__(self, language: str = "python", **kwargs):
        super().__init__(**kwargs)
        self.line_area = LineNumberArea(self)
        self.blockCountChanged.connect(self.update_line_area_width)
        self.updateRequest.connect(self.update_line_area)
        self.set_font(gui.Font.mono(as_qt=True))
        self.update_line_area_width(0)
        self.set_current_line_color(gui.Color(128, 128, 128, 20))
        self.set_syntaxhighlighter(language)

    #     self.delimiter = "    "
    #     self.completion_state = 0
    #     self.completing = False
    #     self.cursorPositionChanged.connect(self.reset_completion)

    # def keyPressEvent(self, event):
    #     match event.key():
    #         case constants.Key.Key_Backtab if self.textCursor().hasSelection():
    #             start_cursor = self.get_textCursor()
    #             with start_cursor.edit_block():
    #                 start_pos = start_cursor.selectionStart()
    #                 start_cursor.setPosition(start_pos)
    #                 start_cursor.move_position("start_of_line")
    #                 start_cursor.clearSelection()
    #                 end_cursor = self.get_text_cursor()
    #                 end_pos = end_cursor.selectionEnd()
    #                 end_cursor.setPosition(end_pos)
    #                 end_cursor.move_position("start_of_line")
    #                 delimit_len = len(self.delimiter)
    #                 while start_cursor.anchor() != end_cursor.position():
    #                     start_cursor.move_position(
    #                         "next_character", mode="keep", n=delimit_len
    #                     )
    #                     if start_cursor.selectedText() == self.delimiter:
    #                         start_cursor.removeSelectedText()
    #                     start_cursor.move_position("next_block")
    #                 start_cursor.move_position(
    #                     "next_character", mode="keep", n=delimit_len
    #                 )
    #                 if start_cursor.selectedText() == self.delimiter:
    #                     start_cursor.removeSelectedText()
    #         case constants.Key.Key_Tab if self.textCursor().hasSelection():
    #             start_cursor = self.get_text_cursor()
    #             with start_cursor.edit_block():
    #                 start_pos = start_cursor.selectionStart()
    #                 start_cursor.setPosition(start_pos)
    #                 start_cursor.move_position("start_of_line")
    #                 end_cursor = self.get_text_cursor()
    #                 end_pos = end_cursor.selectionEnd()
    #                 end_cursor.setPosition(end_pos)
    #                 end_cursor.move_position("start_of_line")
    #                 while start_cursor.position() != end_cursor.position():
    #                     start_cursor.insertText(self.delimiter)
    #                     start_cursor.move_position("next_block")
    #                 start_cursor.insertText(self.delimiter)
    #         case constants.Key.Key_Escape if self.completion_state > 0:
    #             self.completion_state = 0
    #             cursor = self.get_text_cursor()
    #             with cursor.edit_block():
    #                 self.selecter.replace_block_at_cursor(self._orig_text)
    #             self._orig_text == None
    #         case constants.Key.Key_Tab:
    #             if self.is_start():
    #                 self.textCursor().insertText(self.delimiter)
    #             else:
    #                 cursor = self.get_text_cursor()
    #                 with cursor.edit_block():
    #                     self.completing = True
    #                     if self.completion_state == 0:
    #                         self._orig_text = self.textCursor().block().text()
    #                     if self.completion_state > 0:
    #                         self.selecter.replace_block_at_cursor(self._orig_text)
    #                     new_text = self.completer.complete(
    #                         self._orig_text, self.completion_state
    #                     )
    #                     if new_text:
    #                         if new_text.find("(") > 0:
    #                             new_text = new_text[0 : new_text.find("(") + 1]
    #                         self.completion_state += 1
    #                         self.selecter.replace_block_at_cursor(new_text)
    #                     else:
    #                         self.completion_state = 0
    #                         self.selecter.replace_block_at_cursor(self._orig_text)
    #                         self._orig_text == None
    #                 self.completing = False
    #         case _:
    #             return super().keyPressEvent(event)

    # def reset_completion(self):
    #     if not self.completing:
    #         self.completion_state = 0

    # def is_start(self) -> bool:
    #     temp_cursor = self.textCursor()
    #     if temp_cursor.positionInBlock() == 0:
    #         return True
    #     start_text = temp_cursor.block().text()[0 : temp_cursor.positionInBlock()]
    #     delim = set(self.delimiter)
    #     return set(start_text) - delim == set()

    def resizeEvent(self, event):
        super().resizeEvent(event)

        cr = self.contentsRect()
        rect = core.Rect(cr.left(), cr.top(), self.line_area_width(), cr.height())
        self.line_area.setGeometry(rect)

    def line_area_width(self) -> int:
        digits = len(str(self.blockCount()))
        return 3 + self.fontMetrics().boundingRect("9").width() * digits

    def update_line_area_width(self, _):
        self.setViewportMargins(self.line_area_width(), 0, 0, 0)

    def update_line_area(self, rect: core.QRect, dy: int):
        if dy:
            self.line_area.scroll(0, dy)
        else:
            self.line_area.update(0, rect.y(), self.line_area.width(), rect.height())

        if rect.contains(self.viewport().rect()):
            self.update_line_area_width(0)

    def line_area_paintevent(self, event):
        with gui.Painter(self.line_area) as painter:
            painter.fill_rect(event.rect(), "lightgray")

            block = self.firstVisibleBlock()
            block_number = block.blockNumber()
            top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
            bottom = top + self.blockBoundingRect(block).height()
            width = self.line_area.width()
            height = self.fontMetrics().height()
            painter.set_color("black")
            while block.isValid() and (top <= event.rect().bottom()):
                if block.isVisible() and (bottom >= event.rect().top()):
                    number = str(block_number + 1)
                    painter.drawText(
                        0, int(top), width, height, constants.ALIGN_RIGHT, number
                    )
                block = block.next()
                top = bottom
                bottom = top + self.blockBoundingRect(block).height()
                block_number += 1


class LineNumberArea(widgets.Widget):
    def __init__(self, editor: CodeEditor):
        super().__init__(editor)
        self.editor = editor

    def sizeHint(self) -> core.Size:
        return core.Size(self.editor.line_area_width(), 0)

    def paintEvent(self, event):
        self.editor.line_area_paintevent(event)


if __name__ == "__main__":
    app = widgets.app()
    dlg = CodeEditor()
    dlg.show()
    app.exec()



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from prettyqt import constants, gui, itemmodels


if TYPE_CHECKING:
    from prettyqt import core


def escaped(text: str, entity_type: str | None = None) -> str:
    """Helper function to escape telegram markup symbols.

    Args:
        text: The text.
        entity_type: For the entity types ``PRE``, ``CODE`` and the link
                     part of ``TEXT_LINKS``, only certain characters need to be escaped.
    """
    if entity_type in ["pre", "code"]:
        escape_chars = r"\`"
    elif entity_type == "text_link":
        escape_chars = r"\)"
    else:
        escape_chars = r"_*[]()~`>#+-=|{}.!"

    return re.sub(f"([{re.escape(escape_chars)}])", r"\\\1", text)


logger = logging.getLogger(__name__)

# HTML <h{i}> to font px size map
FONT_SIZE_MAP = {1: 34, 2: 30, 3: 24, 4: 20, 5: 18, 6: 16}


class SliceToMarkdownProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model which transforms cell contents to markdown.

    Mainly used for documentation files.
    Text is formatted based on FontRole, Checkstate role, etc.
    """

    ID = "to_markdown"
    ICON = "mdi.palette-outline"

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not self.indexer_contains(index):
            return super().data(index, role)
        if role != constants.DISPLAY_ROLE:
            return None
        # self.strip_styling = True
        # if role != constants.DISPLAY_ROLE and self.strip_styling:
        #     return None
        label = super().data(index, constants.DISPLAY_ROLE)
        checkstate = super().data(index, constants.CHECKSTATE_ROLE)
        # if not label and checkstate is None:
        #     return ""
        label = escaped(str(label) if label is not None else "")
        if label:
            # background = super().data(index, constants.BACKGROUND_ROLE)
            foreground = super().data(index, constants.FOREGROUND_ROLE)
            if isinstance(foreground, gui.QColor):
                label = f'<span style="color:{foreground.name()}">{label}</span>'
            font = super().data(index, constants.FONT_ROLE)
            if font and font.bold():
                label = f"**{label}**"
            if font and font.italic():
                label = f"*{label}*"
        match checkstate:
            case True | constants.CheckState.Checked | 2:
                # :black_square_button:
                label = f"☑ {label}"
            case False | constants.CheckState.Unchecked | 0:
                # :heavy_check_mark:
                label = f"☐ {label}"
        return label

    # def is_using_display_role(self):
    #     return True

    # def use_display_role(self, val: bool):
    #     pass

    # use_display_role = core.Property(bool, is_using_display_role, use_display_role)


if __name__ == "__main__":
    from prettyqt import debugging, itemdelegates, itemmodels, widgets

    app = widgets.app()

    model = itemmodels.QObjectPropertiesModel(app)
    table = widgets.TableView()
    table.set_model(model)
    # table.proxifier[:, :].color_categories()
    font = gui.QFont()
    font.setBold(True)
    table.proxifier[:, :].style(font=font, foreground="red")
    table.proxifier.get_proxy("to_markdown")
    table.set_size_adjust_policy("content")
    table.setWindowTitle("Color values")
    table.set_icon("mdi.palette")
    # table.show()
    table.adjustSize()
    delegate = itemdelegates.MarkdownItemDelegate(mode="commonmark")
    widget = debugging.ProxyComparerWidget(table.model(), delegate=None)

    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import enum
import re
import re._constants as sre_constants
from typing import TYPE_CHECKING, ClassVar

from prettyqt import core, custom_widgets, widgets


if TYPE_CHECKING:
    from re import Pattern


class RegexFlag(enum.IntEnum):
    NOFLAG = 0
    ASCII = 256  # assume ascii "locale"
    IGNORECASE = 2  # ignore case
    LOCALE = 4  # assume current 8-bit locale
    UNICODE = 32  # assume unicode "locale"
    MULTILINE = 8  # make anchors look for newline
    DOTALL = 16  # make dot match newline
    VERBOSE = 64  # ignore whitespace and comments
    # sre extensions (experimental, don't rely on these)
    TEMPLATE = 1  # unknown purpose, deprecated
    DEBUG = 128  # dump pattern after compilation


class RegexInput(widgets.Widget):
    LABEL_MAP: ClassVar = dict(
        multiline="MultiLine",
        ignorecase="Ignore case",
        ascii="ASCII-only matching",
        dotall="Dot matches newline",
        verbose="Ignore whitespace",
    )
    VALUE_MAP: ClassVar = dict(
        multiline=re.MULTILINE,
        ignorecase=re.IGNORECASE,
        ascii=re.ASCII,
        dotall=re.DOTALL,
        verbose=re.VERBOSE,
    )
    Flags = RegexFlag
    core.Enum(Flags)

    value_changed = core.Signal(object)

    def __init__(
        self,
        show_flags: bool = True,
        show_error: bool = True,
        object_name: str = "regex_input",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        self.set_layout("grid", margin=0)
        self.label_error = widgets.Label()
        self.label_error.set_color("highlight_role")
        self.lineedit = custom_widgets.RegexLineEdit()
        val = self.lineedit.validator
        val.error_occured.connect(self.label_error.set_text)
        val.pattern_updated.connect(self.value_changed)
        self.tb_flags = custom_widgets.BoolDictToolButton(
            text="Flags", icon="mdi.flag-variant-outline"
        )
        self.box[0, 0:1] = self.lineedit
        if show_flags:
            self.box[0, 2] = self.tb_flags
        if show_error:
            self.box[1, 0:2] = self.label_error
        self.tb_flags.value_changed.connect(self._on_value_change)

        self.tb_flags.set_dict(self.LABEL_MAP)

    def _on_value_change(self):
        try:
            val = self.get_value()
        except sre_constants.error:
            return
        else:
            self.value_changed.emit(val)

    def get_pattern(self) -> str:
        return self.lineedit.text()

    def set_pattern(self, value: str):
        self.lineedit.set_text(value)

    def get_flags(self) -> int:
        ret_val = self.Flags(0)
        for identifier, flag in self.VALUE_MAP.items():
            if self.tb_flags[identifier]:
                ret_val |= flag
        return ret_val

    def set_flags(self, value: int):
        for identifier, flag in self.VALUE_MAP.items():
            self.tb_flags[identifier] = bool(value & flag)

    def set_value(self, value: str | Pattern | None):
        match value:
            case None:
                self.set_pattern("")
                self.set_flags(self.Flags(0))
            case str():
                self.set_pattern(value)
                self.set_flags(self.Flags(0))
            case _:
                self.set_pattern(value.pattern)
                self.set_flags(self.Flags(value.flags))

    def get_value(self) -> Pattern:
        return re.compile(self.get_pattern(), self.get_flags())

    def is_valid(self) -> bool:
        return self.lineedit.is_valid()

    pattern = core.Property(
        str,
        get_pattern,
        set_pattern,
        user=True,
        doc="Current pattern as text",
    )
    # flags = core.Property(Flags, get_flags, set_flags)


if __name__ == "__main__":
    app = widgets.app()
    widget = RegexInput(show_flags=True, show_error=True)
    table = widgets.TableView()
    table.set_delegate("editor")
    table.set_model_for(widget)
    table.show()
    widget.show()
    widget.value_changed.connect(print)
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import constants, core, widgets
from prettyqt.qt import QtMultimediaWidgets


# QtMultimediaWidgets.QGraphicsVideoItem.__bases__ = (
#     widgets.GraphicsObject,
#     multimedia.MediaBindableInterface,
# )


class GraphicsVideoItem(
    widgets.GraphicsObjectMixin, QtMultimediaWidgets.QGraphicsVideoItem
):
    def get_offset(self) -> core.PointF:
        return core.PointF(self.offset())

    def get_native_size(self) -> core.SizeF:
        return core.SizeF(self.nativeSize())

    def get_size(self) -> core.SizeF:
        return core.SizeF(self.size())

    def set_aspect_ratio_mode(
        self, mode: constants.AspectRatioModeStr | constants.AspectRatioMode
    ):
        """Set the aspect ratio mode.

        Args:
            mode: aspect ratio mode
        """
        self.setAspectRatioMode(constants.ASPECT_RATIO_MODE.get_enum_value(mode))

    def get_aspect_ratio_mode(self) -> constants.AspectRatioModeStr:
        """Return current aspect ratio mode.

        Returns:
            aspect ratio mode
        """
        return constants.ASPECT_RATIO_MODE.inverse[self.aspectRatioMode()]



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.qt import QtMultimedia
from prettyqt.utils import bidict


ERROR = bidict(
    no_error=QtMultimedia.QScreenCapture.Error.NoError,
    internal_error=QtMultimedia.QScreenCapture.Error.InternalError,
    capturing_not_supported=QtMultimedia.QScreenCapture.Error.CapturingNotSupported,
    capture_failed=QtMultimedia.QScreenCapture.Error.CaptureFailed,
    not_found=QtMultimedia.QScreenCapture.Error.NotFound,
)

ErrorStr = Literal["none", "resource", "format", "network", "access_denied"]


class ScreenCapture(core.ObjectMixin, QtMultimedia.QScreenCapture):
    def get_error(self) -> ErrorStr:
        """Return error type.

        Returns:
            error type
        """
        return ERROR.inverse[self.error()]

    def set_screen(self, screen: int | gui.QScreen | Literal["primary"]):
        match screen:
            case int():
                screen = gui.GuiApplication.screens()[screen]
            case gui.QScreen():
                pass
            case "primary":
                screen = gui.GuiApplication.primaryScreen()
            case _:
                raise TypeError(screen)
        self.setScreen(screen)


if __name__ == "__main__":
    player = ScreenCapture()
    source = player.get_source()
    print(source.isValid())



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.utils import colors, get_repr


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class ColorComboBox(widgets.ComboBox):
    value_changed = core.Signal(gui.QColor)

    def __init__(
        self,
        color: datatypes.ColorType | None = None,
        object_name: str = "color_combobox",
        **kwargs,
    ):
        super().__init__(object_name=object_name, **kwargs)
        self._current_color: gui.QColor = gui.QColor("white")
        for i in gui.Color.colorNames():
            self.addItem(iconprovider.for_color(i), i)
        if color is not None:
            self.set_current_color(color)
        self.currentTextChanged.connect(self.set_current_color)

    def __repr__(self):
        return get_repr(self, self._current_color)

    def clear(self):
        self._current_color = gui.QColor("white")
        super().clear()
        for i in gui.Color.colorNames():
            self.addItem(iconprovider.for_color(i), i)

    def set_current_color(self, color: datatypes.ColorType):
        self._current_color = colors.get_color(color).as_qt()
        for color_name in gui.Color.colorNames():
            if gui.Color(color_name) == self._current_color:
                self.setCurrentText(color_name)
                return

    def is_valid(self) -> bool:
        return self._current_color.isValid()

    def get_value(self) -> gui.QColor:
        return self._current_color

    def set_value(self, value: datatypes.ColorType):
        self.set_current_color(value)

    current_color = core.Property(
        gui.QColor,
        get_value,
        set_value,
        user=True,
        doc="Currently selected color",
    )


if __name__ == "__main__":
    app = widgets.app()
    btn = ColorComboBox(gui.Color("green"))
    btn.show()
    btn.value_changed.connect(print)
    app.exec()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Literal

from prettyqt import core, widgets


if TYPE_CHECKING:
    from collections.abc import Iterator


class FlagSelectionWidget(widgets.GroupBox):
    value_changed = core.Signal(int)

    def __init__(
        self,
        label: str = "",
        layout: Literal["horizontal", "vertical"] = "vertical",
        object_name: str = "flag_selection_widget",
        **kwargs,
    ):
        super().__init__(title=label, object_name=object_name, **kwargs)
        self.set_layout(layout)
        self.buttons: dict[widgets.CheckBox, int] = {}

    def __iter__(self) -> Iterator[tuple[widgets.CheckBox, int]]:
        return iter(self.buttons.items())

    def add_items(self, items: Iterable | Mapping):
        if isinstance(items, Mapping):
            for k, v in items.items():
                self.add(v, k)
        else:
            for i in items:
                if isinstance(i, Iterable):
                    self.add(*i)
                else:
                    msg = "Invalid item type"
                    raise TypeError(msg)

    def add(self, title: str, flag: int):
        checkbox = widgets.CheckBox(title, toggled=self.update_choice)
        self.buttons[checkbox] = flag
        self.box.add(checkbox)

    def current_choice(self) -> int:
        ret_val = 0
        for btn, flag in self.buttons.items():
            if btn.isChecked():
                ret_val |= flag
        return int(ret_val)

    @core.Slot(bool)
    def update_choice(self, checked: bool):
        choice = self.current_choice()
        self.value_changed.emit(choice)

    def set_value(self, value: int):
        for btn, flag in self.buttons.items():
            btn.setChecked(bool(value & flag))

    def get_value(self) -> int:
        return self.current_choice()


if __name__ == "__main__":
    import re

    app = widgets.app()
    widget = FlagSelectionWidget()
    items = {re.MULTILINE: "MultiLine", re.IGNORECASE: "Ignore case"}
    widget.add_items(items)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import gui
from prettyqt.utils import get_repr


class TextFrameMixin(gui.TextObjectMixin):
    def __repr__(self):
        return get_repr(self)

    def get_first_cursor_position(self) -> gui.TextCursor:
        return gui.TextCursor(self.firstCursorPosition())

    def get_last_cursor_position(self) -> gui.TextCursor:
        return gui.TextCursor(self.lastCursorPosition())


class TextFrame(TextFrameMixin, gui.QTextFrame):
    pass


if __name__ == "__main__":
    doc = gui.TextDocument()
    frame = TextFrame(doc)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, iconprovider, widgets
from prettyqt.utils import datatypes


if TYPE_CHECKING:
    from prettyqt import gui


class IconWidget(widgets.Label):
    def __init__(self, *names, parent: widgets.QWidget | None = None, **kwargs):
        super().__init__(parent=parent)
        self._icon: gui.QIcon | None = None
        self._size = core.Size(16, 16)
        self.set_icon(iconprovider._icon(*names, **kwargs))

    def set_icon(self, _icon: datatypes.IconType):
        """Set a new icon().

        Parameters
        ----------
        _icon: qtawesome.icon
            icon to set
        """
        self._icon = iconprovider.get_icon(_icon)
        self.setPixmap(self._icon.pixmap(self._size))

    def set_icon_size(self, size: datatypes.SizeType):
        self._size = datatypes.to_size(size)
        self.update()

    def update(self, *args, **kwargs):
        if self._icon:
            self.setPixmap(self._icon.pixmap(self._size))
        return super().update(*args, **kwargs)


if __name__ == "__main__":
    app = widgets.app()
    widget = IconWidget()
    widget.show()
    bool(widget)
    app.exec()



from __future__ import annotations

import logging

from prettyqt import core, eventfilters, itemmodels, widgets


logger = logging.getLogger(__name__)


class QObjectPropertiesTableView(widgets.TableView):
    def __init__(self, *args, **kwargs):
        self.event_catcher = None
        self._handles: list[core.QMetaObject.Connection] = []
        super().__init__(*args, **kwargs)
        self.set_icon("mdi.folder")
        self.set_selection_behavior("rows")
        self.setEditTriggers(self.EditTrigger.AllEditTriggers)
        self.set_delegate("editor", column=1)

    def set_qobject(
        self,
        qobject: core.QObject,
        update_on_event: bool = True,
        update_on_signal_emission: bool = True,
    ):
        prev_model = self.get_model(skip_proxies=True)
        if prev_model is not None and prev_model._qobject:
            self.unhook()
        model = itemmodels.QObjectPropertiesModel(qobject, parent=self)
        # model.dataChanged.connect(self.repaint)
        self.set_model(model)
        if update_on_event:
            self.event_catcher = eventfilters.EventCatcher(
                include=["resize", "move"], parent=qobject
            )
            logger.debug("Connected %r to %r", qobject, model)
            self.event_catcher.caught.connect(model.force_layoutchange)
            qobject.installEventFilter(self.event_catcher)
        if update_on_signal_emission:
            metaobj = core.MetaObject(qobject.metaObject())
            self._handles = metaobj.connect_signals(
                qobject, model.force_layoutchange, only_notifiers=True
            )

    def unhook(self):
        model = self.get_model(skip_proxies=True)
        for handle in self._handles:
            model._qobject.disconnect(handle)
        model._qobject.removeEventFilter(self.event_catcher)
        logger.debug("Disconnected %r from %r", model._qobject, self)


if __name__ == "__main__":
    app = widgets.app()
    with app.debug_mode():
        w = QObjectPropertiesTableView()
        w.set_qobject(app)
        w.show()
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import widgets
from prettyqt.utils import datatypes, listdelegators


if TYPE_CHECKING:
    from collections.abc import Iterator


class GraphicsLayoutMixin(widgets.GraphicsLayoutItemMixin):
    def __getitem__(
        self, index: int | slice
    ) -> widgets.QGraphicsItem | listdelegators.ListDelegator[widgets.QGraphicsItem]:
        match index:
            case int():
                if index >= self.count():
                    raise IndexError(index)
                layoutitem = self.itemAt(index)
                return layoutitem.graphicsItem()
            case slice():
                stop = index.stop or self.count()
                rng = range(index.start or 0, stop, index.step or 1)
                items = [self.itemAt(i).graphicsItem() for i in rng]
                return listdelegators.ListDelegator(items)
            case _:
                raise TypeError(index)

    def __setitem__(self, index: int, value: widgets.QGraphicsItem):
        layoutitem = self.itemAt(index)
        layoutitem.setGraphicsItem(value)

    def __delitem__(self, index: int):
        self.removeAt(index)

    def __iter__(self) -> Iterator[widgets.QGraphicsItem]:
        return iter(self[i] for i in range(self.count()))

    def __contains__(self, item):
        return item in self.get_children()

    def __len__(self):
        # for PySide2
        return self.count()

    def get_children(self) -> listdelegators.ListDelegator[widgets.QGraphicsItem]:
        items = [self.itemAt(i).graphicsItem() for i in range(self.count())]
        return listdelegators.ListDelegator(items)

    def set_margin(self, margin: datatypes.MarginsType):
        m = datatypes.to_margins(margin)
        self.setContentsMargins(m.left(), m.top(), m.right(), m.bottom())


class GraphicsLayout(GraphicsLayoutMixin, widgets.QGraphicsLayout):
    """The base class for all layouts in Graphics View."""


if __name__ == "__main__":
    layout = GraphicsLayout()
